{*------------------------------------------------------------------------------
 TFannNetwork encapsulates the Fast Artificial Neural Network.

 fann.sourceforge.net

 @author Mauricio Pereira Maia
 @email mauriciocpa@gmail.com
 @version 2.1

 2015_03_04 : Changed by Blue Hell:

   - not a componenent anymore
   - ANN_SIGMOID_SYMMETRIC is the default activation function for both outputs and internal nodes
   - unique exception type EFannNetwork
   - enumerated type conversions done with typecasts - there was an error on one of the conversions too
   - string type conversionss to AnsiString and PAnsiChar are internal now
   - changed layout

 -------------------------------------------------------------------------------}

unit FannNetwork;

interface

  uses

    SysUtils,
    Classes,
    Fann;

  type

    EFannNetwork = class( Exception);

    CardinalArray  = array[ 0 .. 32768] of Cardinal;  // Array of Cardinal Type
    PCardinalArray =^CardinalArray;                   // Pointer to CardinalArray Type

    TTrainingAlgorithm = (
      taFANN_TRAIN_INCREMENTAL,
      taFANN_TRAIN_BATCH,
      taFANN_TRAIN_RPROP,
      taFANN_TRAIN_QUICKPROP
    );

    TActivationFunction = (
      afFANN_LINEAR,
      afFANN_THRESHOLD,
      afFANN_THRESHOLD_SYMMETRIC,
      afFANN_SIGMOID,
      afFANN_SIGMOID_STEPWISE,
      afFANN_SIGMOID_SYMMETRIC,
      afFANN_SIGMOID_SYMMETRIC_STEPWISE,
      afFANN_GAUSSIAN,
      afFANN_GAUSSIAN_SYMMETRIC,
      afFANN_GAUSSIAN_STEPWISE,
      afFANN_ELLIOT,
      afFANN_ELLIOT_SYMMETRIC,
      afFANN_LINEAR_PIECE,
      afFANN_LINEAR_PIECE_SYMMETRIC,
      afFANN_SIN_SYMMETRIC,
      afFANN_COS_SYMMETRIC,
      afFANN_SIN,
      afFANN_COS
    );

    {*------------------------------------------------------------------------------
     TFannNetwork
     -------------------------------------------------------------------------------}
    TFannNetwork = class
    private
      FFANN                     : PFann;
      FBuilt                    : Boolean;
      FLayers                   : TStrings;
      FLearningRate             : Single;
      FConnectionRate           : Single;
      FLearningMomentum         : Single;
      FActivationFunctionHidden : Cardinal;
      FActivationFunctionOutput : Cardinal;
      FTrainingAlgorithm        : Cardinal;
    private
      procedure   SetLayers( const Value: TStrings);
      procedure   SetConnectionRate( const Value: Single);
      function    GetConnectionRate: Single;
      procedure   SetLearningRate( const Value: Single);
      function    GetLearningRate: Single;
      procedure   SetLearningMomentum( const Value: Single);
      function    GetLearningMomentum: Single;
      procedure   SetTrainingAlgorithm( Value: TTrainingAlgorithm);
      function    GetTrainingAlgorithm: TTrainingAlgorithm;
      procedure   SetActivationFunctionHidden( Value: TActivationFunction);
      function    GetActivationFunctionHidden: TActivationFunction;
      procedure   SetActivationFunctionOutput( Value: TActivationFunction);
      function    GetActivationFunctionOutput: TActivationFunction;
      function    GetMSE: Single;
      function    EnumActivationFunctionToValue( Value: TActivationFunction): Cardinal;
      function    ValueActivationFunctionToEnum( Value: Cardinal): TActivationFunction;
      function    EnumTrainingAlgorithmToValue( Value: TTrainingAlgorithm): Cardinal;
      function    ValueTrainingAlgorithmToEnum( Value: Cardinal): TTrainingAlgorithm;
    public
      constructor Create;
      destructor  Destroy; override;
      procedure   Build;
      procedure   UnBuild;
      function    Train( Input: array of fann_type; Output: array of fann_type): Single;
      procedure   TrainOnFile( FileName: string; MaxEpochs: Cardinal; DesiredError: Single);
      procedure   Run( Inputs: array of fann_type;var Outputs: array of fann_type);
      procedure   SaveToFile( FileName: string);
      procedure   LoadFromFile( FileName: string);

      {*------------------------------------------------------------------------------
       Pointer to the Fann object.
       If you need to call the fann library directly and skip the Delphi Component.
       -------------------------------------------------------------------------------}
      property    FannObject : PFann   read FFANN;
      property    Built      : Boolean read FBuilt;
    public

      {*------------------------------------------------------------------------------
       Network Layer Structure. Each line need to have the number of neurons
       of the layer.
       2
       4
       1
       Will make a three layered network with 2 input neurons, 4 hidden neurons
       and 1 output neuron.
       -------------------------------------------------------------------------------}
      property    Layers: TStrings read FLayers write SetLayers;

      {*------------------------------------------------------------------------------
       Network Learning Rate.
       -------------------------------------------------------------------------------}
      property    LearningRate: Single read GetLearningRate write SetLearningRate;

      {*------------------------------------------------------------------------------
       Network Connection Rate. See the FANN docs for more info.
       -------------------------------------------------------------------------------}
      property    ConnectionRate: Single read GetConnectionRate write SetConnectionRate;

      {*------------------------------------------------------------------------------
       Network Learning Momentum. See the FANN docs for more info.
       -------------------------------------------------------------------------------}
      property    LearningMometum: Single read GetLearningMomentum write SetLearningMomentum;

      {*------------------------------------------------------------------------------
       Fann Network Mean Square Error. See the FANN docs for more info.
       -------------------------------------------------------------------------------}
      property    MSE: Single read GetMSE;

      {*------------------------------------------------------------------------------
       Training Algorithm used by the network. See the FANN docs for more info.
       -------------------------------------------------------------------------------}
      property    TrainingAlgorithm: TTrainingAlgorithm read GetTrainingAlgorithm write SetTrainingAlgorithm;

      {*------------------------------------------------------------------------------
       Activation Function used by the hidden layers. See the FANN docs for more info.
       -------------------------------------------------------------------------------}
      property    ActivationFunctionHidden: TActivationFunction read GetActivationFunctionHidden write SetActivationFunctionHidden;

      {*------------------------------------------------------------------------------
       Activation Function used by the output layers. See the FANN docs for more info.
       -------------------------------------------------------------------------------}
      property    ActivationFunctionOutput: TActivationFunction read GetActivationFunctionOutput write SetActivationFunctionOutput;
    end;



implementation


  {*------------------------------------------------------------------------------
   Builds a Fann neural nerwork.
   Layers property must be set to work.
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.Build;
  var
    l  : PCardinalArray; // Pointer to number of neurons on each layer
    nl : integer;        // Number of layers
    i  : integer;
  begin
    nl := FLayers.Count;

    if nl = 0
    then EFannNetwork.Create( 'You need to specify the Layers property');

    GetMem( l, nl * sizeof( Cardinal)); // Alloc mem for the array

    for i := 0 to nl - 1
    do l[ i]:= StrtoInt( FLayers.Strings[ i]);

    if FConnectionRate < 1
    then FFANN := fann_create_sparse_array( FConnectionRate, nl, @ l[ 0])
    else FFANN := fann_create_standard_array( nl, @ l[ 0]);

    fann_randomize_weights( FFANN, - 0.5, 0.5);
    fann_set_learning_rate( FFANN, FLearningRate);
    fann_set_learning_momentum( FFANN, FLearningMomentum);
    fann_set_activation_steepness_hidden( FFANN, 0.5);
    fann_set_activation_steepness_output( FFANN, 0.5);
    fann_set_activation_function_hidden( FFANN, FActivationFunctionHidden);
    fann_set_activation_function_output( FFANN, FActivationFunctionOutput);
    fann_set_training_algorithm( FFANN, FTrainingAlgorithm);
    FBuilt := True;
    FreeMem( l);
  end;

  {*------------------------------------------------------------------------------
   Creates an instance of TFannNetwork

   @param      AOwner Owner Component
   -------------------------------------------------------------------------------}
  constructor TFannNetwork.Create;
  begin
    FFANN                     := nil;
    FBuilt                    := False;
    FConnectionRate           := 1;
    FLearningRate             := 0.7;
    FLayers                   := TStringList.Create;
    FLearningMomentum         := 0;
    FActivationFunctionHidden := FANN_SIGMOID_SYMMETRIC;
    FActivationFunctionOutput := FANN_SIGMOID_SYMMETRIC;
    FTrainingAlgorithm        := FANN_TRAIN_RPROP;
  end;

  {*------------------------------------------------------------------------------
   Destroys the TFannNetwork object and frees its memory.
   -------------------------------------------------------------------------------}
  destructor  TFannNetwork.Destroy;
  begin
    UnBuild;
    FreeAndNil( FLayers);
    inherited;
  end;

  {*------------------------------------------------------------------------------
   Gets the activation function used by the hidden layers

   @return Activation Function
   -------------------------------------------------------------------------------}
  function    TFannNetwork.GetActivationFunctionHidden: TActivationFunction;
  begin
    Result := ValueActivationFunctionToEnum( FActivationFunctionHidden);
  end;

  {*------------------------------------------------------------------------------
   Gets the activation function used by the output layer

   @return Activation Function
   -------------------------------------------------------------------------------}
  function    TFannNetwork.GetActivationFunctionOutput: TActivationFunction;
  begin
    Result := ValueActivationFunctionToEnum( FActivationFunctionOutput);
  end;

  {*------------------------------------------------------------------------------
   Gets the network connection rate

   @return Connection Rate
   -------------------------------------------------------------------------------}
  function    TFannNetwork.GetConnectionRate: Single;
  begin
    if Assigned( FFANN)
    then Result := fann_get_connection_rate( FFANN)
    else Result := FConnectionRate;
  end;

  {*------------------------------------------------------------------------------
   Gets the network learning momentum

   @return Learning Momentum
   -------------------------------------------------------------------------------}
  function    TFannNetwork.GetLearningMomentum: Single;
  begin
    if Assigned( FFANN)
    then Result := fann_get_learning_momentum( FFANN)
    else Result := FLearningMomentum;
  end;

  {*------------------------------------------------------------------------------
   Gets the network learning rate

   @return Learning Rate
   -------------------------------------------------------------------------------}
  function    TFannNetwork.GetLearningRate: Single;
  begin
    if Assigned( FFANN)
    then Result := fann_get_learning_rate( FFANN)
    else Result := FLearningRate;
  end;

  {*------------------------------------------------------------------------------
   Returns the Mean Square Error of the Network.

   @return             Mean Square Error
   -------------------------------------------------------------------------------}
  function    TFannNetwork.GetMSE: Single;
  begin
    if not FBuilt
    then raise EFannNetwork.Create( 'The nework has not been built')
    else Result := fann_get_mse( FFANN);
  end;

  {*------------------------------------------------------------------------------
   Gets the network training algorithm

   @return Training algorithm
   -------------------------------------------------------------------------------}
  function    TFannNetwork.GetTrainingAlgorithm: TTrainingAlgorithm;
  begin
    if Assigned( FFANN)
    then Result := ValueTrainingAlgorithmToEnum( fann_get_training_algorithm( FFANN))
    else Result := ValueTrainingAlgorithmToEnum( FTrainingAlgorithm);
  end;

  {*------------------------------------------------------------------------------
   Loads a network from a file.

   @param  Filename    File that contains the network
   @see                SavetoFile
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.LoadFromFile( FileName: string);
  begin
    if FBuilt
    then fann_destroy( FFANN);

    FFANN  := fann_create_from_file( PAnsiChar( AnsiString( FileName)));
    FBuilt := True;
  end;

  {*------------------------------------------------------------------------------
   Executes the network.

   @param  Inputs      Array with the value of each input
   @param  Ouputs      Will receive the output of the network. Need to be
   allocated before the Run function call with the number
   of outputs on the network.
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.Run( Inputs: array of fann_type;var Outputs: array of fann_type);
  var
    O : Pfann_type_array;
    i : integer;
  begin
    if not FBuilt
    then raise EFannNetwork.Create( 'The nework has not been built');

    O := fann_run( FFANN, @ Inputs[ 0]);

    for i := 0 to high( Outputs)
    do Outputs[ i]:= O[ i];
end;

  {*------------------------------------------------------------------------------
   Saves the current network to the specified file.

   @param  Filename    Filename of the network

   @see                LoadFromFile
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SaveToFile( FileName: string);
  begin
    if not FBuilt
    then exit;
    fann_save( FFANN, PAnsiChar( AnsiString( FileName)));
  end;

  {*------------------------------------------------------------------------------
   Sets the activation function for the hidden layers

   @param  Value       The new activation function
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SetActivationFunctionHidden( Value: TActivationFunction);
  begin
    if Assigned( FFANN)
    then fann_set_activation_function_hidden( FFANN, EnumActivationFunctionToValue( Value));

    FActivationFunctionHidden := EnumActivationFunctionToValue( Value);
  end;

  {*------------------------------------------------------------------------------
   Sets the activation function for the output layer

   @param  Value       The activation function
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SetActivationFunctionOutput( Value: TActivationFunction);
  begin
    if Assigned( FFANN)
    then fann_set_activation_function_output( FFANN, EnumActivationFunctionToValue( Value));

    FActivationFunctionOutput := EnumActivationFunctionToValue( Value);
  end;

  {*------------------------------------------------------------------------------
   Sets the connection rate

   @param  Value       The new connection rate
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SetConnectionRate( const Value: Single);
  begin
    FConnectionRate := Value;
  end;

  {*------------------------------------------------------------------------------
   Set the layer structure of the neural network.
   Each line corresponds to a layers on the neural network.

   @param  Value      Text with the layers specification
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SetLayers( const Value: TStrings);
  begin
    if Value.Count < 2
    then raise EFannNetwork.Create( 'The network should have at least two layers')
    else FLayers.Assign( Value);
  end;

  {*------------------------------------------------------------------------------
   Sets the network learning momentum

   @param  Value       The new learning momentum
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SetLearningMomentum( const Value: Single);
  begin
    if Assigned( FFANN)
    then fann_set_learning_momentum( FFANN, Value);
    FLearningMomentum := Value;
  end;

  {*------------------------------------------------------------------------------
   Sets the network learning rate

   @param  Value       The new learning rate
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SetLearningRate( const Value: Single);
  begin
    if Assigned( FFANN)
    then fann_set_learning_rate( FFANN, Value);

    FLearningRate := Value;
  end;

  {*------------------------------------------------------------------------------
   Sets the network training algorithm

   @param  Value       The new training algorithm
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.SetTrainingAlgorithm( Value: TTrainingAlgorithm);
  begin
    if Assigned( FFANN)
    then fann_set_training_algorithm( FFANN, EnumTrainingAlgorithmToValue( Value));

    FTrainingAlgorithm := EnumTrainingAlgorithmToValue( Value);
  end;

  {*------------------------------------------------------------------------------
   Will train one iteration with a set of inputs, and a set of desired outputs.

   @param  Input       Array with the value of each input of the network
   @param  Output      Array with the value of each desired output of the network
   @return             Mean Square Error after the training

   @see                TrainOnFile
   -------------------------------------------------------------------------------}
  function    TFannNetwork.Train( Input, Output: array of fann_type): Single;
  begin
    if not FBuilt
    then raise EFannNetwork.Create( 'The nework has not been built');

    fann_reset_mse( FFANN);
    fann_train( FFANN, @ Input[ 0], @ Output[ 0]);

    Result := fann_get_mse( FFANN);
  end;

  {*------------------------------------------------------------------------------
   Trains the network using the data in filename until desired error is reached
   or until maxepochs is surpassed.

   @param  FileName    Training Data File.
   @param  MaxEpochs   Maximum number of epochs to train
   @param  DesiredError Desired Error of the network after the training

   @see                Train
   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.TrainOnFile( FileName: string; MaxEpochs: Cardinal; DesiredError: Single);
  begin
    if not FBuilt
    then raise EFannNetwork.Create( 'The nework has not been built');

    fann_train_on_file( FFANN, PAnsiChar( AnsiString( FileName)), MaxEpochs, 1000, DesiredError);
  end;

  {*------------------------------------------------------------------------------
   Converts the delphi enum type to the fann define type.

   -------------------------------------------------------------------------------}
  function    TFannNetwork.EnumActivationFunctionToValue( Value: TActivationFunction): Cardinal;
  begin
    Result := Ord( Value);
  end;

  {*------------------------------------------------------------------------------
   Converts the fann activation value to the delphi enum type

   -------------------------------------------------------------------------------}
  function    TFannNetwork.ValueActivationFunctionToEnum( Value: Cardinal): TActivationFunction;
  begin
    Result := TActivationFunction( Value);
  end;

  {*------------------------------------------------------------------------------
   Converts the delphi enum training algorithm to the fann type

   -------------------------------------------------------------------------------}
  function    TFannNetwork.EnumTrainingAlgorithmToValue( Value: TTrainingAlgorithm): Cardinal;
  begin
    Result := Ord( Value);
  end;

  {*------------------------------------------------------------------------------
   Converts the fann training algorithm to the delphi enum type

   -------------------------------------------------------------------------------}
  function    TFannNetwork.ValueTrainingAlgorithmToEnum( Value: Cardinal): TTrainingAlgorithm;
  begin
    Result := TTrainingAlgorithm( Value);
  end;

  {*------------------------------------------------------------------------------
   Destroys the Fann neural nerwork.

   -------------------------------------------------------------------------------}
  procedure   TFannNetwork.UnBuild;
  begin
    if FBuilt
    then fann_destroy( FFANN);

    FBuilt := False;
  end;

end.
