unit TransparentCanvas;

{
  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.

  The Original Code is an alpha-aware canvas class and associated classes.

  The Initial Developer of the Original Code is David Millington.
  Portions created by David Millington are Copyright (C) 2008-2012.
  All Rights Reserved.

  Contributor(s): David Millington.
                  Frank Staal
}

{
  article  : http://parnassus.co/transparent-graphics-with-pure-gdi-part-2-and-introducing-the-ttransparentcanvas-class/
  download : https://github.com/vintagedave/transparent-canvas

  2015-05-01 Adapted for use in Wren by Blue Hell
}

interface

uses

  Windows, SysUtils, Classes, Controls, Graphics, Types

{$if CompilerVersion >= 23.0} // XE2
  , UITypes                   // Let inline function TFont.GetStyle be expanded
{$ifend}

  ;


type

  ETransparentCanvasException = class( Exception);

  TQuadColor = record
    constructor Create( aColor : TColor);
    procedure   Clear;
    function    WrittenByGDI : Boolean;
    procedure   SetAlpha( const aTransparency : Byte; const aPreMult : Single);
    function    AsColorRef : COLORREF;
    procedure   SetFromColorRef      ( const aValue : COLORREF  );
    procedure   SetFromColorMultAlpha( const aValue : TQuadColor);

    case Boolean of
      // These values are not in the same order as COLORREF RGB values - don't assign from a COLORREF directly
      True : (
        Blue  : Byte;
        Green : Byte;
        Red   : Byte;
        Alpha : Byte;
      );
      False : (
        Quad : Cardinal
      );
  end;

  PQuadColor  = ^TQuadColor;
  PPQuadColor = ^PQuadColor;

  TGDIObjects = record
  private
    FBrush : HBRUSH;
    FPen   : HPEN;
    FFont  : HFONT;
  public
    constructor CreateWithHandles( const hBrush : HBRUSH; const hPen : HPEN; const hFont : HFONT);
  public
    property Brush : HBRUSH read FBrush;
    property Pen   : HPEN   read FPen;
    property Font  : HFONT  read FFont;
  end;

  TAlphaBitmapWrapper = class( TPersistent)
  private
    FDCHandle        : HDC;
    FBMPHandle       : HBitmap;
    FOriginalBMP     : HBitmap;
    FQuads           : PQuadColor;
    FWidth           : Integer;
    FHeight          : Integer;
    FOriginalObjects : TGDIObjects;
  private
    procedure   Construct( aDC: HDC; anEmptyFlag: Boolean; aWidth, aHeight: Integer; anInvesionFlag : boolean = false);
    procedure   CheckHandles;
    procedure   Clear;
  public
    constructor CreateBlank( aDC: HDC; aWidth, aHeight: Integer);
    // The DummyX parameters are to avoid duplicate constructors with the same parameter list being
    // inaccessible from C++, although only necessary if you write C++ code that uses this ( internal)
    // class
    constructor CreateForGDI( aDC: HDC; aWidth, aHeight: Integer; aDummyGDI : Byte = 0);
    constructor CreateForDrawThemeTextEx( aDC: HDC; aWidth, aHeight: Integer; aDummyDrawThemeTextEx : SmallInt = 0);
    constructor Create( var aSrc : TAlphaBitmapWrapper);
    destructor  Destroy; override;
  public
    procedure   SelectObjects( const aGDIObjs : TGDIObjects);
    procedure   SelectOriginalObjects;
  public
    procedure   SetAllTransparency ( const anAlpha: Byte);
    procedure   ProcessTransparency( const anAlpha: Byte);                                                     overload;
    procedure   ProcessTransparency( const anAlpha: Byte; aTranspRect : TRect);                                overload;
    procedure   ProcessMaskTransparency( var aMaskImage: TAlphaBitmapWrapper);
    procedure   ProcessTransparentColor( const aTransparentColor : COLORREF; const aTransparentEdgeWidth : Integer = -1);
    procedure   TintByAlphaToColor( const aColor : TQuadColor);
    procedure   BlendTo( anX, anY: Integer; var anImage: TAlphaBitmapWrapper; aTransparency: Byte = $FF);
    procedure   BlendToStretch( anX, anY, aStretchWidth, StretchHeight: Integer; var anImage: TAlphaBitmapWrapper; aTransparency: Byte);
    procedure   BlendToDC( anX, anY : Integer; aDC : HDC; aTransparency : Byte = $FF);
  public
    function    GetRawPixelPtr( const anX, anY : Integer) : PQuadColor;
    procedure   SafeSetRawPixel( const anX, anY : Integer; aColor : TQuadColor);
  published
    function    QuadPointer  : PQuadColor;
  published
    property    Handle       : HDC     read FDCHandle;
    property    BitmapHandle : HBitmap read FBMPHandle;
    property    Width        : Integer read FWidth;
    property    Height       : Integer read FHeight;
  end;

  TCustomTransparentCanvas = class( TPersistent)
    class function TColorToQuadColor( aValue : TColor    ) : TQuadColor;
    class function QuadColorToTColor( aValue : TQuadColor) : TColor;
  private
    FFont       : TFont;
    FBrush      : TBrush;
    FPen        : TPen;
    FAttachedDC : HDC;
  private
    function GetHandle( ) : HDC;
  private
    procedure   SetFont( NewFont : TFont);
    procedure   SetBrush( NewBrush : TBrush);
    procedure   SetPen( NewPen : TPen);
  private
    function    GetPenPos : TPoint;
    procedure   SetPenPos( NewPos : TPoint);
  private
    function    GetWidth : Integer;
    function    GetHeight : Integer;
  private
    // Converts to non-premultiplied alpha
    function    GetPixel( X, Y : Integer) : COLORREF;
    procedure   SetPixel( X, Y: Integer; Color: Cardinal); overload;
    procedure   SetPixel( X, Y: Integer; Color: Cardinal; Alpha: Byte); overload;
  private
    // Direct pre-multiplied alpha
    function    GetRawPixel( X, Y : Integer) : TQuadColor;
    procedure   SetRawPixel( X, Y : Integer; Color : TQuadColor);
  private
    procedure   TextOutPreVista ( const aRect : TRect; const aText: string; const anAlignment : TAlignment; const anAlpha : Byte);
    procedure   TextOutVistaPlus( const aRect : TRect; const aText: string; const anAlignment : TAlignment; const anAlpha : Byte);
    function    CanUseDrawThemeTextEx : boolean;
    procedure   InternalGlowTextOut(
      const anX, anY, aGlowSize   : Integer;
      const aText                 : string;
      const anAlignment           : TAlignment;
      const anAlpha               : Byte;
      const aProcessBackColorFlag : Boolean;
      const aBackColor            : TQuadColor
    ); overload;
    procedure InternalGlowTextOut(
      const aRect                 : TRect;
      const aGlowSize             : Integer;
      const aText                 : string;
      const anAlignment           : TAlignment;
      const anAlpha               : Byte;
      const aProcessBackColorFlag : Boolean;
      const aBackColor            : TQuadColor
    ); overload;
  protected
    FWorkingCanvas : TAlphaBitmapWrapper;
  protected
    function    OrphanAliasedFont : HFONT;
  public
    constructor Create( aWidth, aHeight : Integer);                                                            overload;
    constructor Create( aCanvas: TCanvas);                                                                     overload;
    constructor Create( aDC : HDC; aWidth, aHeight : Integer);                                                 overload;
    constructor Create( aSrc : TCustomTransparentCanvas);                                                      overload;
    destructor  Destroy;                                                                                       override;
  public
    procedure   Assign  ( aSource : TPersistent);                                                              override;
    procedure   AssignTo( aDest   : TPersistent);                                                              override;
  public
    procedure   SaveToFile( const aFilename : string);
  public
    procedure   Draw(
      const anX, anY              : Integer;
            aCanvas               : TCanvas;
      const aWidth, aHeight       : Integer;
      const UseTransparentColor   : Boolean  = false;
      const aTransparentColor     : COLORREF = $0;
      const aTransparentEdgeWidth : Integer  = -1
    );                                                                                                         overload;
    procedure   Draw(
      const anX, anY        : Integer;
      const aMetafile       : TMetafile;
      const aWidth, aHeight : Integer;
      const aTransparency   : Byte = $FF
    );                                                                                                         overload;
    procedure   Draw(
      const anX, anY      : Integer;
            aTarget       : TCustomTransparentCanvas;
      const aTransparency : Byte = 255
    );                                                                                                         overload;
  public
    procedure   DrawTo(
      const anX, anY                    : Integer;
            aCanvas                     : TCanvas;
      const aTargetWidth, aTargetHeight : Integer;
      const aTransparency               : Byte = $FF
    );                                                                                                         overload;
    procedure   DrawTo(
      const anX, anY                    : Integer;
            aDC                         : HDC;
      const aTargetWidth, aTargetHeight : Integer;
      const aTransparency               : Byte = $FF
    );                                                                                                         overload;
    procedure   DrawToGlass( const anX, anY : Integer; aDC : HDC; const aTransparency : Byte = $FF);
  public
    procedure   Ellipse( const anX1, anY1, anX2, anY2: Integer; const anAlpha : Byte = $FF);                   overload;
    procedure   Ellipse( const aRect : TRect;                   const anAlpha : Byte = $FF);                   overload;
  public
    procedure   MoveTo( const anX, anY: Integer);
  public
    procedure   RoundRect( const anX1, anY1, anX2, anY2, anXRadius, anYRadius: Integer; const anAlpha : Byte = $FF); overload;
    procedure   RoundRect( const aRect : TRect;    const anXRadius, anYRadius: Integer; const anAlpha : Byte = $FF); overload;
    procedure   Rectangle( const anX1, anY1, anX2, anY2: Integer;                       const anAlpha : Byte = $FF); overload;
    procedure   Rectangle( const aRect : TRect;                                         const anAlpha : Byte = $FF); overload;
  public
    function    TextExtent( const aText: string): TSize;
    function    TextHeight( const aText: string): Integer;
    procedure   TextOut ( const anX, anY: Integer; const aText: string; const anAlignment : TAlignment = taLeftJustify; const anAlpha : Byte = $FF);
    procedure   TextRect( const aRect: TRect;      const aText: string; const anAlignment : TAlignment = taLeftJustify; const anAlpha : Byte = $FF);
    function    TextWidth( const aText: string): Integer;
  public
    function    CanDrawGlowText : boolean;
    procedure   GlowTextOut(
      const anX, anY, aGlowSize : Integer;
      const aText               : string;
      const anAlignment         : TAlignment = taLeftJustify;
      const anAlpha             : Byte = $FF
    );
    procedure GlowTextOutBackColor(
      const anX, anY, aGlowSize : Integer;
      const aText               : string;
      const aBackColor          : TColor;
      const anAlignment         : TAlignment = taLeftJustify;
      const aGlowAlpha          : Byte = $FF;
      const anAlpha             : Byte = $FF
    );
  public
    procedure   Clear;
  public
    property    Handle                    : HDC        read GetHandle;
    property    PenPos                    : TPoint     read GetPenPos   write SetPenPos;
    property    Pixels[X, Y: Integer]     : COLORREF   read GetPixel    write SetPixel;
    property    RawPixels[ X, Y: Integer] : TQuadColor read GetRawPixel write SetRawPixel;
  published
    property    Brush                     : TBrush     read FBrush      write SetBrush;
    property    Font                      : TFont      read FFont       write SetFont;
    property    Pen                       : TPen       read FPen        write SetPen;
    property    Width                     : Integer    read GetWidth;
    property    Height                    : Integer    read GetHeight;
  end;

  TTransparentCanvas = class( TCustomTransparentCanvas);

  TTransparentControlCanvas = class( TCustomTransparentCanvas)
  private
    FControl   : TWinControl;
    FControlDC : HDC;
  public
    constructor Create( aControl : TWinControl);
    destructor  Destroy;                                                                                       override;
  end;



implementation



uses

  Math, Themes, UxTheme, RTLConsts;


{$if CompilerVersion >= 23.0} // XE2

  function InternalStyleServices : TCustomStyleServices;
  begin

    {$if declared( StyleServices)}

      Result := StyleServices;

    {$else}

      Result := ThemeServices; // Deprecated in favour of StyleServices

    {$ifend}

  end;

{$else}

  function InternalStyleServices : TThemeServices;
  begin
    Result := ThemeServices;
  end;

{$ifend}

  function AlignmentToFlags( const Alignment : TAlignment) : DWORD;
  begin
    Result := 0;
    case Alignment of
      taLeftJustify: Result := DT_LEFT;
      taRightJustify: Result := DT_RIGHT;
      taCenter: Result := DT_CENTER;
    end;
  end;


  { ========
    TCustomTransparentCanvas
  }

    function    TCustomTransparentCanvas.CanDrawGlowText: boolean;
    begin
      Result := CanUseDrawThemeTextEx;
    end;

    function    TCustomTransparentCanvas.CanUseDrawThemeTextEx: boolean;
    begin
      Result :=
        {$if declared( StyleServices)} // Can't test TCustomStyleServices.Enabled, assume deprecation follows StyleServices
          InternalStyleServices.Enabled
        {$else}
          InternalStyleServices.ThemesEnabled
        {$ifend}
        and ( Win32MajorVersion >= 6);
    end;

    procedure   TCustomTransparentCanvas.Clear;
    begin
      FWorkingCanvas.Clear;
    end;

    constructor TCustomTransparentCanvas.Create( aWidth, aHeight : Integer);
    begin
      inherited Create;
      FWorkingCanvas := TAlphaBitmapWrapper.CreateBlank( 0, aWidth, aHeight);
      FFont          := TFont.Create;
      FBrush         := TBrush.Create;
      FPen           := TPen.Create;
      FAttachedDC    := 0;
    end;

    constructor TCustomTransparentCanvas.Create( aSrc: TCustomTransparentCanvas);
    begin
      inherited Create;
      FWorkingCanvas := TAlphaBitmapWrapper.Create( aSrc.FWorkingCanvas);

      FFont := TFont.Create;
      FFont.Assign( aSrc.FFont);

      FBrush := TBrush.Create;
      FBrush.Assign( aSrc.FBrush);

      FPen := TPen.Create;
      FPen.Assign( aSrc.FPen);

      FAttachedDC := 0;
    end;

    constructor TCustomTransparentCanvas.Create( aCanvas: TCanvas);
    begin
      inherited Create;
      FAttachedDC    := aCanvas.Handle;
      FWorkingCanvas := TAlphaBitmapWrapper.CreateBlank( aCanvas.Handle, Width, Height);
      FFont          := TFont.Create;
      FBrush         := TBrush.Create;
      FPen           := TPen.Create;
    end;

    constructor TCustomTransparentCanvas.Create( aDC: HDC; aWidth, aHeight : Integer);
    begin
      inherited Create;
      FAttachedDC    := aDC;
      FWorkingCanvas := TAlphaBitmapWrapper.CreateBlank( aDC, aWidth, aHeight);
      FFont          := TFont.Create;
      FBrush         := TBrush.Create;
      FPen           := TPen.Create;
    end;

    destructor  TCustomTransparentCanvas.Destroy;
    begin
      FreeAndNil( FWorkingCanvas);
      FreeAndNil( FFont);
      FreeAndNil( FBrush);
      FreeAndNil( FPen);

      inherited;
    end;

    procedure   TCustomTransparentCanvas.Assign( aSource : TPersistent);
    var
      ToCopy : TCustomTransparentCanvas;
    begin
      if aSource is TCustomTransparentCanvas
      then begin
        ToCopy := aSource as TCustomTransparentCanvas;

        FWorkingCanvas.Free;
        FWorkingCanvas := TAlphaBitmapWrapper.Create( ToCopy.FWorkingCanvas);

        assert( Assigned( FFont) and Assigned( FBrush) and Assigned( FPen));

        FFont .Assign( ToCopy.FFont );
        FBrush.Assign( ToCopy.FBrush);
        FPen  .Assign( ToCopy.FPen  );

        FAttachedDC := 0;
      end
      else begin
        if Assigned( aSource)
        then raise EConvertError.CreateResFmt( @RTLConsts.SAssignError, [ aSource.ClassName, ClassName])
        else raise EConvertError.CreateResFmt( @RTLConsts.SAssignError, [ 'nil', ClassName]);
      end;
    end;

    procedure   TCustomTransparentCanvas.AssignTo( aDest: TPersistent);
    begin
      aDest.Assign( Self);
    end;

    procedure   TCustomTransparentCanvas.SaveToFile( const aFilename : string);
    var
      BMP : TBitmap;
    begin
      // Draw to a transparent 32-bit bitmap, and save that
      BMP := TBitmap.Create;
      try
        BMP.PixelFormat := pf32bit;
        BMP.Width       := Width;
        BMP.Height      := Height;

        DrawTo( 0, 0, BMP.Canvas, Width, Height);

        BMP.SaveToFile( aFilename);
      finally
        BMP.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.Draw(
      const anX, anY: Integer;
      aCanvas: TCanvas;
      const aWidth, aHeight : Integer;
      const UseTransparentColor : Boolean;
      const aTransparentColor : COLORREF;
      const aTransparentEdgeWidth : Integer
    );
    var
      TempImage : TAlphaBitmapWrapper;
    begin
      TempImage := TAlphaBitmapWrapper.CreateForGDI( FWorkingCanvas.FDCHandle, aWidth, aHeight);
      try
        BitBlt( TempImage.FDCHandle, 0, 0, aWidth, aHeight, aCanvas.Handle, 0, 0, SRCCOPY);
        TempImage.SetAllTransparency( $FF); // No need to test, all written by GDI

        if UseTransparentColor
        then TempImage.ProcessTransparentColor( aTransparentColor, aTransparentEdgeWidth);

        TempImage.BlendTo( anX, anY, FWorkingCanvas);
      finally
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.Draw(
      const anX, anY: Integer;
      const aMetafile: TMetafile;
      const aWidth, aHeight: Integer;
      const aTransparency : Byte = $FF
    );
    var
      TempImage : TAlphaBitmapWrapper;
    begin
      TempImage := TAlphaBitmapWrapper.CreateForGDI( FWorkingCanvas.FDCHandle, aWidth, aHeight);
      try
        TempImage.SelectObjects( TGDIObjects.CreateWithHandles( Brush.Handle, Pen.Handle, Font.Handle));
        try
          PlayEnhMetaFile( TempImage.FDCHandle, aMetafile.Handle, Rect( 0, 0, aWidth, aHeight));
          TempImage.ProcessTransparency( aTransparency);
          TempImage.BlendTo( anX, anY, FWorkingCanvas);
        finally
          TempImage.SelectOriginalObjects;
        end;
      finally
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.Draw(
      const anX, anY : Integer;
      aTarget : TCustomTransparentCanvas;
      const aTransparency : Byte = 255
    );
    begin
      aTarget.FWorkingCanvas.BlendTo( anX, anY, FWorkingCanvas, aTransparency);
    end;

    procedure   TCustomTransparentCanvas.DrawTo(
      const anX, anY: Integer;
      aCanvas: TCanvas;
      const aTargetWidth, aTargetHeight: Integer;
      const aTransparency : Byte = 255
    );
    begin
      DrawTo( anX, anY, aCanvas.Handle, aTargetWidth, aTargetHeight, aTransparency);
    end;

    procedure   TCustomTransparentCanvas.DrawTo(
      const anX, anY: Integer;
      aDC: HDC;
      const aTargetWidth, aTargetHeight: Integer;
      const aTransparency : Byte = 255
    );
    var
      TempCanvas: TAlphaBitmapWrapper;
    begin
      // Create a 32-bit canvas with a copy of the dc drawn in it with opaque alpha
      TempCanvas := TAlphaBitmapWrapper.CreateBlank( aDC, aTargetWidth, aTargetHeight);
      try
        BitBlt( TempCanvas.FDCHandle, 0, 0, aTargetWidth, aTargetHeight, aDC, 0, 0, SRCCOPY);
        TempCanvas.SetAllTransparency( $FF);

        // Now blend the working image onto it at ( X, Y), possibly stretched
        if ( aTargetWidth = Width) and ( aTargetHeight = Height)
        then FWorkingCanvas.BlendTo( anX, anY, TempCanvas, aTransparency)
        else FWorkingCanvas.BlendToStretch( anX, anY, aTargetWidth, aTargetHeight, TempCanvas, aTransparency);

        // Now blit the composited image back to the DC
        BitBlt( aDC, 0, 0, aTargetWidth, aTargetHeight, TempCanvas.FDCHandle, 0, 0, SRCCOPY);
      finally
        TempCanvas.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.DrawToGlass( const anX, anY: Integer; aDC: HDC; const aTransparency : Byte);
    begin
      FWorkingCanvas.BlendToDC( anX, anY, aDC, aTransparency);
    end;

    procedure   TCustomTransparentCanvas.Ellipse( const anX1, anY1, anX2, anY2: Integer; const anAlpha: Byte);
    var
      TempImage : TAlphaBitmapWrapper;
    begin
      TempImage := TAlphaBitmapWrapper.CreateForGDI( FWorkingCanvas.FDCHandle, anX2-anX1, anY2-anY1);
      try
        TempImage.SelectObjects( TGDIObjects.CreateWithHandles( Brush.Handle, Pen.Handle, Font.Handle));
        SetWindowOrgEx( TempImage.FDCHandle, anX1 - Pen.Width div 2, anY1 - Pen.Width div 2, nil);
        Windows.Ellipse( TempImage.FDCHandle, anX1, anY1, anX2, anY2);
        SetWindowOrgEx( TempImage.FDCHandle, 0, 0, nil);
        TempImage.ProcessTransparency( anAlpha);
        TempImage.BlendTo( anX1, anY1, FWorkingCanvas);
        TempImage.SelectOriginalObjects;
      finally
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.Ellipse( const aRect: TRect; const anAlpha: Byte);
    begin
      Ellipse( aRect.Left, aRect.Top, aRect.Right, aRect.Bottom, anAlpha);
    end;

    function    TCustomTransparentCanvas.GetHandle: HDC;
    begin
      Result := FWorkingCanvas.FDCHandle;
    end;

    function    TCustomTransparentCanvas.GetHeight: Integer;
    begin
      Result := FWorkingCanvas.FHeight;
    end;

    function    TCustomTransparentCanvas.GetPenPos: TPoint;
    begin
      GetCurrentPositionEx( FWorkingCanvas.FDCHandle, @Result);
    end;

    function    TCustomTransparentCanvas.GetPixel( X, Y: Integer): COLORREF;
    begin
      Result := GetRawPixel( X, Y).AsColorRef;
    end;

    function    TCustomTransparentCanvas.GetRawPixel( X, Y: Integer): TQuadColor;
    var
      PQuad    : PQuadColor;
      InverseY : Integer;
    begin
      InverseY := FWorkingCanvas.FHeight - Y - 1; // 0 is top not bottom
      PQuad    := FWorkingCanvas.FQuads;
      Inc( PQuad, ( InverseY * Width) + X);
      Result := PQuad^;
    end;

    function    TCustomTransparentCanvas.GetWidth: Integer;
    begin
      Result := FWorkingCanvas.FWidth;
    end;

    procedure   TCustomTransparentCanvas.InternalGlowTextOut(
      const anX, anY, aGlowSize: Integer;
      const aText: string;
      const anAlignment : TAlignment;
      const anAlpha: Byte;
      const aProcessBackColorFlag : Boolean;
      const aBackColor : TQuadColor
    );
    var
      TextSize : TSize;
    begin
      TextSize := TextExtent( aText);
      InternalGlowTextOut( Rect( anX, anY, anX + TextSize.cx, anY + TextSize.cy), aGlowSize, aText, anAlignment, anAlpha, aProcessBackColorFlag, aBackColor);
    end;

    procedure  TCustomTransparentCanvas.InternalGlowTextOut(
      const aRect : TRect;
      const aGlowSize: Integer;
      const aText: string;
      const anAlignment : TAlignment;
      const anAlpha: Byte;
      const aProcessBackColorFlag : Boolean;
      const aBackColor : TQuadColor);
    var
      TempImage  : TAlphaBitmapWrapper;
      TextSize   : TSize;
      Options    : TDTTOpts;
      Details    : TThemedElementDetails;
      TextRect   : TRect;
      AlignFlags : DWORD;
    begin
      if Length( aText) = 0
      then Exit; // Crash creating zero-sized bitmap

      if not CanDrawGlowText
      then raise ETransparentCanvasException.Create( 'Cannot use DrawThemeTextEx');

      TextSize := aRect.Size;
      AlignFlags := AlignmentToFlags( anAlignment);
      TempImage := TAlphaBitmapWrapper.CreateForDrawThemeTextEx( FWorkingCanvas.FDCHandle, TextSize.cx + aGlowSize*2, TextSize.cy + aGlowSize*2);

      try
        TempImage.SelectObjects( TGDIObjects.CreateWithHandles( 0, 0, Font.Handle));
        SetBkMode( TempImage.FDCHandle, TRANSPARENT);
        SetTextColor( TempImage.FDCHandle, ColorToRGB( Font.Color));

        ZeroMemory( @Options, SizeOf( Options));
        Options.dwSize := SizeOf( Options);
        Options.dwFlags := DTT_TEXTCOLOR or DTT_COMPOSITED or DTT_GLOWSIZE;
        Options.crText := ColorToRGB( Font.Color);
        Options.iGlowSize := aGlowSize;

        Details  := InternalStyleServices.GetElementDetails( teEditTextNormal);
        TextRect := Rect( aGlowSize, aGlowSize, TextSize.cx + aGlowSize*2, TextSize.cy + aGlowSize*2);

        DrawThemeTextEx(
          InternalStyleServices.Theme[ teEdit],
          TempImage.FDCHandle,
          Details.Part,
          Details.State,
          PChar( aText),
          Length( aText),
          AlignFlags or DT_TOP or DT_NOCLIP,
          TextRect,
          Options
        );

        if aProcessBackColorFlag
        then begin
          TempImage.TintByAlphaToColor( aBackColor);
          // Now draw the text over again, but with no glow, so only the text is drawn
          TextRect := Rect( aGlowSize, aGlowSize, TextSize.cx + aGlowSize, TextSize.cy + aGlowSize);
          Options.dwFlags := DTT_TEXTCOLOR or DTT_COMPOSITED;
          Options.crText := ColorToRGB( Font.Color);
          Options.iGlowSize := 0;
          DrawThemeTextEx(
            InternalStyleServices.Theme[ teEdit],
            TempImage.FDCHandle,
            Details.Part,
            Details.State,
            PChar( aText),
            Length( aText),
            AlignFlags or DT_TOP or DT_NOCLIP,
            TextRect,
            Options
          );
        end;

        case anAlignment of
          taLeftJustify  : TempImage.BlendTo( aRect.Left - aGlowSize, aRect.Top - aGlowSize, FWorkingCanvas, anAlpha);
          taRightJustify : TempImage.BlendTo( aRect.Left - aGlowSize * 2, aRect.Top - aGlowSize, FWorkingCanvas, anAlpha);
          taCenter       : TempImage.BlendTo( aRect.Left - aGlowSize * 2 + 1 + IfThen( aGlowSize > 0, 1, -1), aRect.Top - aGlowSize, FWorkingCanvas, anAlpha);
        end;
        SetBkMode( TempImage.FDCHandle, OPAQUE);
        TempImage.SelectOriginalObjects;
      finally
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.GlowTextOut(
      const anX, anY, aGlowSize: Integer;
      const aText: string;
      const anAlignment : TAlignment;
      const anAlpha: Byte
    );
    begin
      InternalGlowTextOut( anX, anY, aGlowSize, aText, anAlignment, anAlpha, False, TQuadColor.Create( 0));
    end;

    procedure   TCustomTransparentCanvas.GlowTextOutBackColor(
      const anX, anY, aGlowSize: Integer;
      const aText: string;
      const aBackColor: TColor;
      const anAlignment : TAlignment;
      const aGlowAlpha : Byte;
      const anAlpha: Byte
    );
    var
      Background : TQuadColor;
    begin
      if ( COLORREF( ColorToRGB( aBackColor)) = RGB( 255, 255, 255)) and ( aGlowAlpha = 255) // White is the default on
      then GlowTextOut( anX, anY, aGlowSize, aText, anAlignment, anAlpha)                            //Windows; do no special processing
      else begin
        // Windows draws glowing text with a white background, always.  To change the background colour,
        // draw with the normal white background and black text, then process the colours to change
        // white to the specified colour, and black to the font colour
        Background := TQuadColor.Create( aBackColor);
        Background.SetAlpha( aGlowAlpha, anAlpha / 255.0);
        InternalGlowTextOut( anX, anY, aGlowSize, aText, anAlignment, anAlpha, True, Background);
      end;
    end;

    procedure   TCustomTransparentCanvas.MoveTo( const anX, anY: Integer);
    begin
      MoveToEx( FWorkingCanvas.FDCHandle, anX, anY, nil);
    end;

    function    TCustomTransparentCanvas.OrphanAliasedFont: HFONT;
    var
      FontWeight : Cardinal;
    begin
      // Font output and alpha is tricky with a ClearType or antialiased font.  This method takes FFont
      // and creates a new font with the same attributes, but with ClearType and AA explicitly disabled
      if fsBold in Font.Style
      then FontWeight := FW_BOLD
      else FontWeight := FW_NORMAL;

      Result :=
        CreateFont(
          FFont.Height,
          0,
          0,
          0,
          FontWeight,
          Cardinal( fsItalic in Font.Style),
          Cardinal( fsUnderline in Font.Style),
          Cardinal( fsStrikeOut in Font.Style),
          DEFAULT_CHARSET,
          OUT_DEFAULT_PRECIS,
          CLIP_DEFAULT_PRECIS,
          ANTIALIASED_QUALITY,
          DEFAULT_PITCH,
          PChar( Font.Name)
        );
    end;

    class function TCustomTransparentCanvas.QuadColorToTColor( aValue: TQuadColor): TColor;
    begin
      Result := TColor( RGB( aValue.Red, aValue.Blue, aValue.Green));
    end;

    procedure   TCustomTransparentCanvas.Rectangle( const anX1, anY1, anX2, anY2: Integer; const anAlpha: Byte);
    var
      TempImage : TAlphaBitmapWrapper;
    begin
      TempImage := TAlphaBitmapWrapper.CreateForGDI( FWorkingCanvas.FDCHandle, anX2-anX1, anY2-anY1);
      try
        TempImage.SelectObjects( TGDIObjects.CreateWithHandles( Brush.Handle, Pen.Handle, Font.Handle));
        SetWindowOrgEx( TempImage.FDCHandle, anX1 - Pen.Width div 2, anY1 - Pen.Width div 2, nil);
        Windows.Rectangle( TempImage.FDCHandle, anX1, anY1, anX2, anY2);
        SetWindowOrgEx( TempImage.FDCHandle, 0, 0, nil);
        TempImage.ProcessTransparency( anAlpha);
        TempImage.BlendTo( anX1, anY1, FWorkingCanvas);
        TempImage.SelectOriginalObjects;
      finally
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.Rectangle( const aRect: TRect; const anAlpha: Byte);
    begin
      Rectangle( aRect.Left, aRect.Top, aRect.Right, aRect.Bottom, anAlpha);
    end;

    procedure   TCustomTransparentCanvas.RoundRect( const aRect: TRect; const anXRadius, anYRadius : Integer; const anAlpha : Byte = $FF);
    begin
      RoundRect( aRect.Left, aRect.Top, aRect.Right, aRect.Bottom, anXRadius, anYRadius, anAlpha);
    end;

    procedure   TCustomTransparentCanvas.RoundRect( const anX1, anY1, anX2, anY2, anXRadius, anYRadius: Integer; const anAlpha : Byte = $FF);
    var
      TempImage : TAlphaBitmapWrapper;
    begin
      TempImage := TAlphaBitmapWrapper.CreateForGDI( FWorkingCanvas.FDCHandle, anX2-anX1 + Pen.Width, anY2-anY1 + Pen.Width);
      try
        TempImage.SelectObjects( TGDIObjects.CreateWithHandles( Brush.Handle, Pen.Handle, Font.Handle));
        SetWindowOrgEx( TempImage.FDCHandle, anX1 - Pen.Width div 2, anY1 - Pen.Width div 2, nil);
        Windows.RoundRect( TempImage.FDCHandle, anX1, anY1, anX2, anY2, anXRadius, anYRadius);
        SetWindowOrgEx( TempImage.FDCHandle, 0, 0, nil);
        TempImage.ProcessTransparency( anAlpha);
        TempImage.BlendTo( anX1, anY1, FWorkingCanvas);
        TempImage.SelectOriginalObjects;
      finally
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.SetBrush( NewBrush: TBrush);
    begin
      FBrush.Assign( NewBrush);
    end;

    procedure   TCustomTransparentCanvas.SetFont( NewFont: TFont);
    begin
      FFont.Assign( NewFont);
    end;

    procedure   TCustomTransparentCanvas.SetPen( NewPen: TPen);
    begin
      FPen.Assign( NewPen);
    end;

    procedure   TCustomTransparentCanvas.SetPenPos( NewPos: TPoint);
    begin
      MoveToEx( FWorkingCanvas.FDCHandle, NewPos.X, NewPos.Y, nil);
    end;

    procedure   TCustomTransparentCanvas.SetPixel( X, Y: Integer; Color: Cardinal);
    begin
      SetPixel( X, Y, Color, $FF);
    end;

    procedure   TCustomTransparentCanvas.SetPixel( X, Y: Integer; Color: Cardinal; Alpha: Byte);
    var
      PQuad    : PQuadColor;
      InverseY : Integer;
    begin
      InverseY := FWorkingCanvas.FHeight - Y - 1; // 0 is top not bottom
      PQuad    := FWorkingCanvas.FQuads;
      Inc( PQuad, ( InverseY * Width) + X);
      PQuad.Quad  := Color;
      PQuad.Alpha := Alpha;
    end;

    procedure   TCustomTransparentCanvas.SetRawPixel( X, Y: Integer; Color: TQuadColor);
    var
      PQuad    : PQuadColor;
      InverseY : Integer;
    begin
      InverseY := FWorkingCanvas.FHeight - Y - 1; // 0 is top not bottom
      PQuad    := FWorkingCanvas.FQuads;
      Inc( PQuad, ( InverseY * Width) + X);
      PQuad.Quad := Color.Quad;
    end;

    class function TCustomTransparentCanvas.TColorToQuadColor( aValue: TColor): TQuadColor;
    begin
      Result := TQuadColor.Create( aValue);
    end;

    function    TCustomTransparentCanvas.TextExtent( const aText: string): TSize;
    var
      OldFontHandle,
      FontHandle : HFONT;
    begin
      if Length( aText) = 0
      then begin
        Result.cx := 0;
        Result.cy := 0;
        Exit;
      end;

      if CanUseDrawThemeTextEx
      then begin
        // Can use DrawThemeTextEx; just get text extent normally
        FWorkingCanvas.SelectObjects( TGDIObjects.CreateWithHandles( 0, 0, Font.Handle));
        GetTextExtentPoint32( FWorkingCanvas.FDCHandle, PChar( aText), Length( aText), Result);
        FWorkingCanvas.SelectOriginalObjects;
      end
      else begin
        // Can't use DrawThemeTextEx; use aliased font ( may affect output size, so need to explicitly
        // measure using the aliased font)
        FontHandle := OrphanAliasedFont;
        try
          OldFontHandle := SelectObject( FWorkingCanvas.FDCHandle, FontHandle);
          GetTextExtentPoint32( FWorkingCanvas.FDCHandle, PChar( aText), Length( aText), Result);
          SelectObject( FWorkingCanvas.FDCHandle, OldFontHandle);
        finally
          DeleteObject( FontHandle);
        end;
      end;
    end;

    function    TCustomTransparentCanvas.TextHeight( const aText: string): Integer;
    begin
      Result := TextExtent( aText).cy;
    end;

    procedure   TCustomTransparentCanvas.TextOut(
      const anX, anY: Integer;
      const aText: string;
      const anAlignment : TAlignment;
      const anAlpha : Byte
    );
    var
      TextSize : TSize;
    begin
      if Length( aText) = 0           // Crash creating zero-sized bitmap
      then Exit;

      TextSize := TextExtent( aText);
      if CanUseDrawThemeTextEx
      then TextOutVistaPlus( Rect( anX, anY, anX + TextSize.cx, anY + TextSize.cy), aText, anAlignment, anAlpha)
      else TextOutPreVista ( Rect( anX, anY, anX + TextSize.cx, anY + TextSize.cy), aText, anAlignment, anAlpha);
    end;

    procedure   TCustomTransparentCanvas.TextOutPreVista(
      const aRect: TRect;
      const aText: string;
      const anAlignment : TAlignment;
      const anAlpha: Byte
    );
    var
      TempImage  : TAlphaBitmapWrapper;
      FontHandle : HFONT;
      TextSize   : TSize;
      OldAlign   : UINT;
    begin
      if Length( aText) = 0
      then Exit; // Crash creating zero-sized bitmap

      TextSize := TextExtent( aText);
      // Clip to the rest by restricting the size it thinks the text is - the bitmap will be this size, thus clipped
      TextSize.cx := min( TextSize.cx, aRect.Right-aRect.Left);
      TextSize.cy := min( TextSize.cy, aRect.Bottom-aRect.Top);
      FontHandle  := OrphanAliasedFont; // Antialiased or cleartype text works terribly when trying to fix the alpha
      TempImage   := TAlphaBitmapWrapper.CreateForGDI( FWorkingCanvas.FDCHandle, TextSize.cx, TextSize.cy);

      try
        TempImage.SelectObjects( TGDIObjects.CreateWithHandles( 0, 0, FontHandle));
        SetBkMode( TempImage.FDCHandle, TRANSPARENT);
        SetTextColor( TempImage.FDCHandle, ColorToRGB( Font.Color));
        OldAlign := GetTextAlign( TempImage.FDCHandle);

        try
          case anAlignment of
            taLeftJustify  : SetTextAlign( TempImage.FDCHandle, TA_LEFT);
            taRightJustify : SetTextAlign( TempImage.FDCHandle, TA_RIGHT);
            taCenter       : SetTextAlign( TempImage.FDCHandle, TA_CENTER);
          end;

          ExtTextOut( TempImage.FDCHandle, 0, 0, ETO_CLIPPED, nil, PChar( aText), Length( aText), nil);
        finally
          SetTextAlign( TempImage.FDCHandle, OldAlign);
        end;
        SetBkMode( TempImage.FDCHandle, OPAQUE);
        TempImage.ProcessTransparency( anAlpha);
        TempImage.BlendTo( aRect.Left, aRect.Top, FWorkingCanvas);
        TempImage.SelectOriginalObjects;
      finally
        DeleteObject( FontHandle);
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.TextOutVistaPlus(
      const aRect: TRect;
      const aText: string;
      const anAlignment : TAlignment;
      const anAlpha: Byte
    );
    var
      TempImage  : TAlphaBitmapWrapper;
      TextSize   : TSize;
      Options    : TDTTOpts;
      Details    : TThemedElementDetails;
      TextRect   : TRect;
      AlignFlags : DWORD;
    begin
      if Length( aText) = 0
      then Exit; // Crash creating zero-sized bitmap

      if not CanUseDrawThemeTextEx
      then raise ETransparentCanvasException.Create( 'Cannot use DrawThemeTextEx');

      AlignFlags := AlignmentToFlags( anAlignment);
      TextSize   := aRect.Size; //TextExtent( Text);

      // Clip by clipping the size of the rectangle it assumes the text fits in
      TextSize.cx := min( TextSize.cx, aRect.Right-aRect.Left);
      TextSize.cy := min( TextSize.cy, aRect.Bottom-aRect.Top);
      TempImage   := TAlphaBitmapWrapper.CreateForDrawThemeTextEx( FWorkingCanvas.FDCHandle, TextSize.cx, TextSize.cy);

      try

        TempImage.SelectObjects( TGDIObjects.CreateWithHandles( 0, 0, Font.Handle));
        SetBkMode( TempImage.FDCHandle, TRANSPARENT);
        SetTextColor( TempImage.FDCHandle, ColorToRGB( Font.Color));

        ZeroMemory( @Options, SizeOf( Options));
        Options.dwSize := SizeOf( Options);
        Options.dwFlags := DTT_TEXTCOLOR or DTT_COMPOSITED;
        Options.crText := ColorToRGB( Font.Color);
        Options.iGlowSize := 0;

        Details := InternalStyleServices.GetElementDetails( teEditTextNormal);

        TextRect := Rect( 0, 0, TextSize.cx, TextSize.cy);
        DrawThemeTextEx(
          InternalStyleServices.Theme[ teEdit],
          TempImage.FDCHandle,
          Details.Part,
          Details.State,
          PChar( aText),
          Length( aText),
          AlignFlags or DT_TOP, TextRect,
          Options
        );

        SetBkMode( TempImage.FDCHandle, OPAQUE);
        TempImage.BlendTo( aRect.Left, aRect.Top, FWorkingCanvas, anAlpha);
        TempImage.SelectOriginalObjects;
      finally
        TempImage.Free;
      end;
    end;

    procedure   TCustomTransparentCanvas.TextRect( const aRect: TRect; const aText: string; const anAlignment : TAlignment; const anAlpha: Byte);
    begin
      if Length( aText) = 0
      then Exit; // Crash creating zero-sized bitmap

      if CanUseDrawThemeTextEx
      then TextOutVistaPlus( aRect, aText, anAlignment, anAlpha)
      else TextOutPreVista ( aRect, aText, anAlignment, anAlpha);
    end;

    function    TCustomTransparentCanvas.TextWidth( const aText: string): Integer;
    begin
      Result := TextExtent( aText).cx;
    end;




  { ========
    TAlphaBitmapWrapper
  }

    procedure   TAlphaBitmapWrapper.BlendTo( anX, anY: Integer; var anImage: TAlphaBitmapWrapper; aTransparency: Byte);
    var
      BlendFunc : TBlendFunction;
    begin
      with BlendFunc
      do begin
        BlendOp             := AC_SRC_OVER;
        BlendFlags          := 0;
        SourceConstantAlpha := aTransparency; // Normally 255
        AlphaFormat         := AC_SRC_ALPHA;
      end;
      AlphaBlend( anImage.FDCHandle, anX, anY, FWidth, FHeight, FDcHandle, 0, 0, FWidth, FHeight, BlendFunc);
    end;

    procedure   TAlphaBitmapWrapper.BlendToDC( anX, anY: Integer; aDC: HDC; aTransparency: Byte);
    var
      BlendFunc : TBlendFunction;
    begin
      with BlendFunc
      do begin
        BlendOp             := AC_SRC_OVER;
        BlendFlags          := 0;
        SourceConstantAlpha := aTransparency;
        AlphaFormat         := AC_SRC_ALPHA;
      end;
      AlphaBlend( aDC, anX, anY, FWidth, FHeight, FDcHandle, 0, 0, FWidth, FHeight, BlendFunc);
    end;

    procedure   TAlphaBitmapWrapper.BlendToStretch( anX, anY, aStretchWidth, StretchHeight: Integer; var anImage: TAlphaBitmapWrapper; aTransparency: Byte);
    var
      BlendFunc : TBlendFunction;
    begin
      with BlendFunc
      do begin
        BlendOp             := AC_SRC_OVER;
        BlendFlags          := 0;
        SourceConstantAlpha := aTransparency; // Normally 255
        AlphaFormat         := AC_SRC_ALPHA;
      end;
      AlphaBlend( anImage.FDCHandle, 0, 0,aStretchWidth, StretchHeight, FDcHandle, anX, anY, FWidth, FHeight, BlendFunc);
    end;

    procedure   TAlphaBitmapWrapper.CheckHandles;
    begin
      if FDCHandle = 0
      then raise ETransparentCanvasException.Create( 'Cannot create device context');

      if FBMPHandle = 0
      then raise ETransparentCanvasException.Create( 'Cannot create 32-bit bitmap');

      if FQuads = nil
      then raise ETransparentCanvasException.Create( 'Cannot access bitmap bits');
    end;

    procedure   TAlphaBitmapWrapper.Clear;
    begin
      ZeroMemory( FQuads, FWidth * FHeight * SizeOf( TQuadColor));
    end;

    procedure   TAlphaBitmapWrapper.Construct( aDC: HDC; anEmptyFlag: Boolean; aWidth, aHeight: Integer; anInvesionFlag : boolean);
    var
      BMPInfo : TBitmapInfo;
      PQuads  : Pointer;
    begin
      FWidth  := aWidth;
      FHeight := aHeight;

      if ( FWidth <= 0) or ( FHeight <= 0)
      then raise ETransparentCanvasException.Create( 'Invalid size specified; Width and Height must both be greater than zero.');

      FDCHandle := CreateCompatibleDC( aDC);
      ZeroMemory( @BMPInfo, SizeOf( TBitmapInfo));

      with BMPInfo.bmiHeader
      do begin
        biSize  := SizeOf( TBitmapInfo);
        biWidth := FWidth;

        if anInvesionFlag
        then biHeight := -FHeight // For DrawThemeTextEx: requires inverted ( bottom-up) bitmap
        else biHeight :=  FHeight;

        biPlanes      := 1;
        biBitCount    := 32;
        biCompression := BI_RGB;
        biSizeImage   := FWidth * FHeight * SizeOf( TQuadColor);
      end;

      PQuads     := nil;
      FBMPHandle := 0;
      FBMPHandle := CreateDIBSection( FDCHandle, BMPInfo, DIB_RGB_COLORS, PQuads, 0, 0);
      FQuads     := PQuadColor( PQuads);

      CheckHandles;

      FOriginalBMP := SelectObject( FDCHandle, FBMPHandle);
      GdiFlush; // Need to flush before any manipulation of bits

      if anEmptyFlag
      then ZeroMemory( FQuads, aWidth * aHeight * SizeOf( TQuadColor))
      else FillMemory( FQuads, aWidth * aHeight * SizeOf( TQuadColor), $FF);
    end;

    constructor TAlphaBitmapWrapper.Create( var aSrc: TAlphaBitmapWrapper);
    begin
      inherited Create;
      Construct( aSrc.FDCHandle, true, aSrc.FWidth, aSrc.FHeight); // true = init to all zeroes
      aSrc.BlendTo( 0, 0, Self);
    end;

    constructor TAlphaBitmapWrapper.CreateBlank( aDC: HDC; aWidth, aHeight: Integer);
    begin
      inherited Create;
      Construct( aDC, true, aWidth, aHeight); // true = init to all zeroes
    end;

    constructor TAlphaBitmapWrapper.CreateForDrawThemeTextEx( aDC: HDC; aWidth, aHeight: Integer; aDummyDrawThemeTextEx : SmallInt = 0);
    begin
      inherited Create;
      Construct( aDC, true, aWidth, aHeight, true); // init to all zeroes; inverted ( upside down) because DrawThemeTextEx needs it
    end;

    constructor TAlphaBitmapWrapper.CreateForGDI( aDC: HDC; aWidth, aHeight: Integer; aDummyGDI : Byte = 0);
    begin
      inherited Create;
      Construct( aDC, false, aWidth, aHeight); // false = init all bytes to $FF, so can test if written to
    end;

    destructor  TAlphaBitmapWrapper.Destroy;
    begin
      SelectOriginalObjects;
      SelectObject( FDCHandle, FOriginalBMP);
      DeleteObject( FBMPHandle);
      FBMPHandle := 0;
      DeleteObject( FDCHandle);
      FDCHandle := 0;
      inherited;
    end;

    procedure   TAlphaBitmapWrapper.TintByAlphaToColor( const aColor: TQuadColor);
    var
      Loop  : Integer;
      PQuad : PQuadColor;
    begin
      // Change the background colour of glowing text by changing white to BackColor, and black to
      // TextColor.  Alpha remains the same
      GdiFlush; // Need to flush before any manipulation of bits
      PQuad := FQuads;
      for Loop := 0 to FWidth * FHeight - 1
      do begin
        if PQuad.Alpha <> 0
        then PQuad.SetFromColorMultAlpha( aColor); // Sets the colour, and multiplies the alphas together
        Inc( PQuad);
      end;
    end;

    procedure   TAlphaBitmapWrapper.ProcessMaskTransparency( var aMaskImage: TAlphaBitmapWrapper);
    var
      Loop      : Integer;
      PQuad     : PQuadColor;
      PMaskQuad : PQuadColor;
    begin
      if not (( FWidth = aMaskImage.FWidth)) and ( FHeight = aMaskImage.FHeight)
      then raise ETransparentCanvasException.Create( 'Mask images must be identical in size');

      GdiFlush; // Need to flush before any manipulation of bits
      PQuad     := FQuads;
      PMaskQuad := aMaskImage.FQuads;
      for Loop := 0 to FWidth * FHeight - 1
      do begin
        if ( PMaskQuad.Quad and $00FFFFFF) = 0
        then PQuad.SetAlpha( 255, 1.0)
        else PQuad.Quad := 0;

        Inc( PQuad);
        Inc( PmaskQuad);
      end;
    end;

    procedure   TAlphaBitmapWrapper.ProcessTransparentColor( const aTransparentColor : COLORREF; const aTransparentEdgeWidth : Integer);

      function IsEdge( const PixelIndex : Integer) : Boolean;
      var
        X, Y : Integer;
      begin
        if aTransparentEdgeWidth < 0
        then Exit( true); // Entire image should be processed

        // index = ( Y * width) + X ( note Y is inverse)
        Y := PixelIndex div FWidth;
        X := PixelIndex - ( Y * FWidth);
        Result :=
          ( X <             aTransparentEdgeWidth     ) or
          ( Y <             aTransparentEdgeWidth     ) or
          ( X > ( FWidth  - aTransparentEdgeWidth - 1)) or
          ( Y > ( FHeight - aTransparentEdgeWidth - 1));
      end;

    var
      Loop    : Integer;
      PQuad   : PQuadColor;
      R, G, B : Byte;
    begin
      if aTransparentEdgeWidth = 0
      then Exit; // Want to process an edge, but no edge width ( pass -1 for whole image)

      GdiFlush; // Need to flush before any manipulation of bits
      R := GetRValue( aTransparentColor);
      G := GetGValue( aTransparentColor);
      B := GetBValue( aTransparentColor);
      PQuad := FQuads;
      for Loop := 0 to FWidth * FHeight - 1
      do begin
        if ( PQuad.Red = R) and ( PQuad.Green = G) and ( PQuad.Blue = B) and IsEdge( Loop)
        then PQuad.SetAlpha( 0, 0); // 32-bit OSes must have all channels 0 ( premultiplied) for 0 alpha

        Inc( PQuad);
      end;
    end;

    procedure   TAlphaBitmapWrapper.ProcessTransparency( const anAlpha: Byte; aTranspRect: TRect);
    var
      LoopX   : Integer;
      PreMult : Single;
      PQuad   : PQuadColor;
      LoopY   : Integer;
    begin
      GdiFlush; // Need to flush before any manipulation of bits
      IntersectRect( aTranspRect, aTranspRect, Rect( 0, 0, FWidth, FHeight)); // Clip to valid bounds

      PreMult := anAlpha / 255.0;

      for LoopY := aTranspRect.Top to aTranspRect.Bottom - 1
      do begin
        PQuad := FQuads;
        Inc( PQuad, LoopY);

        for LoopX := aTranspRect.Left to aTranspRect.Right - 1
        do begin
          if PQuad.WrittenByGDI
          then PQuad.SetAlpha( anAlpha, PreMult)
          else PQuad.Quad := 0;

          Inc( PQuad);
        end;
      end;
    end;

    procedure   TAlphaBitmapWrapper.ProcessTransparency( const anAlpha: Byte);
    var
      Loop    : Integer;
      PreMult : Single;
      PQuad   : PQuadColor;
    begin
      GdiFlush; // Need to flush before any manipulation of bits
      PreMult := anAlpha / 255.0;
      PQuad   := FQuads;

      for Loop := 0 to FWidth * FHeight - 1
      do begin
        if PQuad.WrittenByGDI
        then PQuad.SetAlpha( anAlpha, PreMult)
        else PQuad.Quad := 0;

        Inc( PQuad);
      end;
    end;

    function    TAlphaBitmapWrapper.QuadPointer: PQuadColor;
    begin
      Result := FQuads;
    end;

    function    TAlphaBitmapWrapper.GetRawPixelPtr( const anX, anY: Integer): PQuadColor;
    var
      PQuad    : PQuadColor;
      InverseY : Integer;
    begin
      if ( anX >= 0) and ( anX < FWidth) and ( anY >= 0) and ( anY < FHeight)
      then begin
        InverseY := FHeight - anY - 1; // 0 is top not bottom
        PQuad    := FQuads;
        Inc( PQuad, ( InverseY * FWidth) + anX);
        Result := PQuad;
      end
      else Result := nil;
    end;

    procedure   TAlphaBitmapWrapper.SafeSetRawPixel( const anX, anY: Integer; aColor: TQuadColor);
    var
      PQuad    : PQuadColor;
      InverseY : Integer;
    begin
      if ( anX >= 0) and ( anX < FWidth) and ( anY >= 0) and ( anY < FHeight)
      then begin
        InverseY := FHeight - anY - 1; // 0 is top not bottom
        PQuad    := FQuads;
        Inc( PQuad, ( InverseY * FWidth) + anX);
        PQuad.Quad := aColor.Quad;
      end;
    end;

    procedure   TAlphaBitmapWrapper.SelectObjects( const aGDIObjs: TGDIObjects);
    begin
      // This is only one layer deep - it stores the old objects in FOriginalObjects
      // If you call it more than once, it will overwrite the values in FOriginalObjects
      // If you find yourself doing this, this needs to be rewritten as a stack, and you'd
      // push and pop the handles.

      if ( FOriginalObjects.FBrush <> 0) or ( FOriginalObjects.FPen <> 0) or ( FOriginalObjects.FFont <> 0)
      then raise ETransparentCanvasException.Create( 'SelectObjects has already been called');

      FOriginalObjects.FBrush := SelectObject( FDCHandle, aGDIObjs.Brush);
      FOriginalObjects.FPen   := SelectObject( FDCHandle, aGDIObjs.Pen);
      FOriginalObjects.FFont  := SelectObject( FDCHandle, aGDIObjs.Font);
    end;

    procedure   TAlphaBitmapWrapper.SelectOriginalObjects;
    begin
      SelectObject( FDCHandle, FOriginalObjects.FBrush); FOriginalObjects.FBrush := 0;
      SelectObject( FDCHandle, FOriginalObjects.FPen  ); FOriginalObjects.FPen   := 0;
      SelectObject( FDCHandle, FOriginalObjects.FFont ); FOriginalObjects.FFont  := 0;
    end;

    procedure   TAlphaBitmapWrapper.SetAllTransparency( const anAlpha: Byte);
    var
      Loop    : Integer;
      PreMult : Single;
      PQuad   : PQuadColor;
    begin
      GdiFlush; // Need to flush before any manipulation of bits
      PreMult := anAlpha / 255.0;
      PQuad   := FQuads;

      for Loop := 0 to FWidth * FHeight - 1
      do begin
        PQuad.SetAlpha( anAlpha, PreMult);
        Inc( PQuad);
      end;
    end;

  { ========
    TQuadColor
  }

    function    TQuadColor.AsColorRef: COLORREF;
    var
      PreDiv : Single;
    begin
      // contains premultiplied alpha, so un-premultiply it and return that
      if Alpha = 0
      then Result := $00000000
      else begin
        PreDiv := 1 / ( Alpha / 255.0);
        Result := RGB( Trunc( Red * PreDiv), Trunc( Green * PreDiv), Trunc( Blue * PreDiv)) or ( Alpha shl 24);
      end;
    end;

    procedure   TQuadColor.Clear;
    begin
      Quad := 0;
    end;

    constructor TQuadColor.Create( aColor: TColor);
    var
      ColorRGB : COLORREF;
    begin
      Alpha    := 255;
      ColorRGB := ColorToRGB( aColor);
      Red      := GetRValue( ColorRGB);
      Green    := GetGValue( ColorRGB);
      Blue     := GetBValue( ColorRGB);
    end;

    procedure   TQuadColor.SetAlpha( const aTransparency: Byte; const aPreMult: Single);
    begin
      Alpha := aTransparency;
      Blue  := Trunc( Blue  * aPreMult);
      Green := Trunc( Green * aPreMult);
      Red   := Trunc( Red   * aPreMult);
    end;

    procedure   TQuadColor.SetFromColorRef( const aValue: COLORREF);
    begin
      // Sets the colour, but keeps the current alpha
      if Alpha = 0
      then Quad := 0
      else begin
        Red   := GetRValue( aValue);
        Green := GetGValue( aValue);
        Blue  := GetBValue( aValue);
        SetAlpha( Alpha, Alpha / 255.0);
      end;
    end;

    procedure   TQuadColor.SetFromColorMultAlpha( const aValue: TQuadColor);
    var
      MultAlpha : Byte;
    begin
      Red       := aValue.Red;
      Green     := aValue.Green;
      Blue      := aValue.Blue;
      MultAlpha := Round( Integer( Alpha) * Integer( aValue.Alpha) / 255.0);
      SetAlpha( MultAlpha, MultAlpha / 255.0);
    end;

    function    TQuadColor.WrittenByGDI: Boolean;
    begin
      Result := ( Alpha = 0);
    end;


  { ========
    TTransparentControlCanvas
  }

    constructor TTransparentControlCanvas.Create( aControl: TWinControl);
    begin
      if aControl = nil
      then raise ETransparentCanvasException.Create( 'Control must not be nil');

      if aControl.Handle = 0
      then raise ETransparentCanvasException.Create( 'Cannot access control handle');

      FControl   := aControl;
      FControlDC := GetWindowDC( aControl.Handle);

      if FControlDC = 0
      then raise ETransparentCanvasException.Create( 'Cannot obtain control device context');

      inherited Create( FControlDC, aControl.Width, aControl.Height);
    end;

    destructor  TTransparentControlCanvas.Destroy;
    begin
      DrawTo( 0, 0, FControlDC, FControl.Width, FControl.Height);

      ReleaseDC( FControl.Handle, FControlDC);
      FControlDC := 0;
      FWorkingCanvas.FDCHandle := 0;

      inherited;
    end;


  { ========
    TGDIObjects
  }

    constructor TGDIObjects.CreateWithHandles( const hBrush: HBRUSH; const hPen: HPEN; const hFont : HFONT);
    begin
      FBrush := hBrush;
      FPen   := hPen;
      FFont  := hFont;
    end;

end.
