unit module_defs;

{

   COPYRIGHT 2013 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

{.$DEFINE PATCH_PROFILER :: best to be defined from build configuration}

interface

uses

  System.SysUtils, System.Math, System.Classes,

  KnobsUtils, KnobsConversions, ComplexMath;

const

  NO_INPUT      = 'noinput';
  NO_OUTPUT     = 'nooutput';
  NO_INTERNAL   = 'nointernal';
  NO_STRING     = 'nostring';
  NO_MODULE     = 'nomodule';
  UNDEF_PMC     =    -1;               // Pin/module/connection index for non existent pin/module/connection
  MAX_INPUTS    =   256;
  MAX_OUTPUTS   =   256;
  MAX_INTERNALS =   256;
  FIFO_SIZE     = 65536;               // Sizes for input and output FIFOs, should be as large as the largest settable size for the audio device queues

  {$IFDEF PATCH_PROFILER}
  HAS_PATCH_PROFILER = True;
  {$ELSE}
  HAS_PATCH_PROFILER = False;
  {$ENDIF}

  WhiteSpace             = [ #$00 .. #$20, #$7f];
  StrictDigits           = [ '0' .. '9'];
  Digits                 = StrictDigits + [ '+', '-'];
  IdentifierStarters     = [ 'A' .. 'Z', 'a' .. 'z', '%', '#', '@', '_'];
  IdentifierChars        = IdentifierStarters + Digits + [ '&'];
  FloatChars             = Digits + [ '.', 'e', 'E'];

  inputtype_normal       = 0;
  inputtype_inverted     = 1;
  inputtype_positive     = 2;
  inputtype_pos_inv      = 3;
  inputtype_negative     = 4;
  inputtype_neg_inv      = 5;

  outputtype_normal      = 0;
  outputtype_inverted    = 1;
  outputtype_positive    = 2;
  outputtype_pos_inv     = 3;
  outputtype_negative    = 4;
  outputtype_neg_inv     = 5;

  envoutputtype_normal   = 0;
  envoutputtype_inverted = 1;
  envoutputtype_negative = 2;
  envoutputtype_neg_inv  = 3;

  // Midi command definitions

  CNoteOff         = $80; // Length = 3;
  CNoteOn          = $90; // Length = 3;
  CKeyPressure     = $a0; // Length = 3;
  CControlChange   = $b0; // Length = 3;
  CProgramChange   = $c0; // Length = 2;
  CChannelPressure = $d0; // Length = 2;
  CPitchBend       = $e0; // Length = 2;
  CSysEx           = $f0; // Length = n;

  // When aCmd And $f0 = CSysEx low nibble of 1st byte is :

  SSysEx           = $00; // Length = 1, followed by arbitrary amt of data, finalized by SEox;
  SSongPosition    = $02; // Length = 3;
  SSongSelect      = $03; // Length = 2;
  STuneRequest     = $06; // Length = 1;
  SEox             = $07; // Length = 1;
  STimingClock     = $08; // Length = 1;
  SStart           = $0a; // Length = 1;
  SContinue        = $0b; // Length = 1;
  SStop            = $0c; // Length = 1;
  SActiveSense     = $0e; // Length = 1;
  SReset           = $0f; // Length = 1;

  CCAllNotesOff    = 123; // Pseudo CC - all notes off

type

  EModule           = class( Exception);
  EModuleNotFound   = class( EModule  );
  EModuleDuplicate  = class( EModule  );
  EInputNotFound    = class( EModule  );
  EOutputNotFound   = class( EModule  );
  EInternalNotFound = class( EModule  );
  EStringNotFound   = class( EModule  );
  EUnitsUndefined   = class( EModule  );
  ENoPinLookup      = class( EModule  );

  TStringArray      = array of string;
  TConnection       = class;
  TConnectionArray  = array of TConnection;
  TMod              = class;
  TSynthPatch       = class;
  TModuleArray      = array of TMod;
  TModuleClass      = class of TMod;

  TModuleIdTypeMap = record
    ModuleId    : Integer;
    ModuleClass : TModuleClass;
  end;

  TModuleIdTypeMaps = array of TModuleIdTypeMap;
  TInts             = array of Integer;
  TKnobMap          = set   of Byte;

  TMidiMessage = record
    Command : Byte;
    Channel : Byte;
    Data    : TInts;
  end;

  TLightsHandler = procedure( const aSender: TMod   ; const aMsg: string)                                of object;
  TDataHandler   = procedure( const aSender: TMod   ; const aPath: string; const aData: TSignalArray   ) of object;
  TXYDataHandler = procedure( const aSender: TMod   ; const aPath: string; const aData: TSignalPairFifo) of object;
  TInfoHandler   = procedure( const aSender: TMod   ; const aMsg: string)                                of object;
  TOnSendMidi    = procedure( const aSender: TObject; const aMsg: TMidiMessage)                          of object;
  TOnSynthLog    = procedure( const aSender: TObject; const aMsg: string)                                of object;
  TOnSynthError  = procedure( const aSender: TObject; const aMsg: string)                                of object;

  TConnection = class
  public
    Peer      : TSynthPatch;
    Index     : Integer;
    SrcModule : Integer;
    SrcPin    : Integer;
    DstModule : Integer;
    DstPin    : Integer;
  private
    procedure   Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
  public
    procedure   Save( anIndent: Integer; var S: string);
    function    IsSameConnection( aConnection: TConnection): Boolean;
  end;

  TConnections = class
  private
    FPeer       : TSynthPatch;
    FConnections: TConnectionArray;
  private
    function    GetCount: Integer;
    function    GetConnection( anIndex: Integer): TConnection;
    procedure   SetConnection( anIndex: Integer; const aValue: TConnection);
  public
    constructor Create( aPeer: TSynthPatch);
    destructor  Destroy;                                                                     override;
    procedure   Save( anIndent: Integer; const aName: string; var S: string);
    procedure   Clear;
    procedure   AddConnection( const aValue: TConnection);                                   overload;
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);           overload;
    procedure   RemoveConnection( anIndex: Integer);
    procedure   RemoveConnectionsFor( aModuleIndex: Integer);
    procedure   PropagateSignals( const aModules: TModuleArray);
    function    IsSameConnections( const aConnections: TConnections): Boolean;
  public
    property    Count: Integer                             read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  end;

  TMod = class
  private
    FParent          : TSynthPatch;   // The patch this module is part off
    FName            : string;        // A unique name for this module - unique within it's parent
    FResetFlag       : Integer;       // Set when a module reset was requested
    FSamplesFast     : Int64;         // Sample counter, incremented each time a new Fast sample is calcultaed
    FSamplesSlow     : Int64;         // Sample counter, incremented each time a new Slow sample is calcultaed
    {$IFDEF PATCH_PROFILER}
  private
    FClocksFast      : Int64;         // CPU Fast clocks for current cycle
    FClocksSlow      : Int64;         // CPU Slow clocks for current cycle
    FTotalClocksFast : Int64;         // CPU Fast clocks accumulated over the module life time
    FTotalClocksSlow : Int64;         // CPU Slow clocks accumulated over the module life time
    FZips            : Int64;         // CPU clocks spent on dezoppering for the current cycle
    FTotalZips       : Int64;         // CPU clocks acumulated FZips over the module life time
    {$ENDIF}
  protected
    FIsFast          : Boolean;       // True when fast mode processing must be performed
    FIsSlow          : Boolean;       // True when slow mode processing must be performed
    FIsMidi          : Boolean;       // True when midi reception must be enabled
    FInputs          : TSignalArray;  // The input signals, these include both patch signals and user controls
    FDezips          : TSignalArray;  // An extra set of inputs for user controls, these are copied to FInCache with some filtering when the signal happens to be be in the deziper map
    FOutputs         : TSignalArray;  // The ouput signals.
    FDezipperMap     : TKnobMap;      // Determines  what signals should be dezippered, most, but not all, knobs should be dezippered
  private
    function    AssertUniqueName( const aName: string): string;                                   virtual;
  private
    procedure   SetName( const aValue: string);
    function    GetResetFlag: Boolean;
    procedure   SetResetFlag( aValue: Boolean);                                                   virtual;
    function    GetInputCount  : Integer;
    function    GetOutputCount : Integer;
    function    GetOutput( anIndex: Integer): TSignal;
    procedure   SetCommonValue        ( anIndex : Integer; aValue: TSignal);                      virtual;
    procedure   SetInternalValue      ( const aName: string; aValue: TSignal);                    virtual;
    procedure   SetInternalStringValue( const aName, aValue: string);
  protected
    procedure   SetStringValue        ( const aName, aValue: string);                             virtual;
    function    GetStringValue        ( const aName: string): string;                             virtual;
    procedure   FixInputs;
    procedure   FixOutputs;
    procedure   SetInputs( aStartIndex: Integer; const aValue: TSignals);
    procedure   HandleParam( const aControl: string; aValue: TSignal);                            virtual;
    procedure   Log( const aMsg: string);                                                         virtual;
    procedure   LogFmt( const aFmt: string; const anArgs: array of const);
    procedure   Error( const aMsg: string);                                                       virtual;
    procedure   ErrorFmt( const aFmt: string; const anArgs: array of const);
    {$IFDEF PATCH_PROFILER}
  protected
    procedure   PreTickFast;                                                                       inline;
    procedure   PreTickSlow;                                                                       inline;
    procedure   PostTickFast;                                                                      inline;
    procedure   PostTickSlow;                                                                      inline;
    procedure   PreZipper;                                                                         inline;
    procedure   PostZipper;                                                                        inline;
    {$ENDIF}
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                               virtual;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                      virtual;
    procedure   SaveInputValues ( anIndent: Integer; var S: string);
    procedure   SaveKeyValue    ( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    procedure   SaveDetails     ( anIndent: Integer; var S: string);                              virtual;
    procedure   SaveInternals   ( anIndent: Integer; var S: string);                              virtual;
    procedure   SetInternal     ( const aName: string; aValue: TSignal);                          virtual;
    function    GetInternal     ( const aName: string): TSignal;                                  virtual;
    procedure   Save            ( anIndent: Integer; var S: string);
    procedure   Reset;                                                                            virtual;
    procedure   ResetSampleCounts;                                                                virtual;
    procedure   Clear;                                                                            virtual;
    function    FindInput   ( const aPinName: string): Integer;
    function    FindOutput  ( const aPinName: string): Integer;
    function    InputName   ( anIndex: Integer): string;                                          virtual;
    function    OutputName  ( anIndex: Integer): string;                                          virtual;
    function    InternalName( anIndex: Integer): string;                                          virtual;
    procedure   Tick;
    procedure   SlowTick;
    procedure   DeZipper;
    procedure   DoTick;                                                                           virtual;
    procedure   DoSlowTick;                                                                       virtual;
    procedure   FixZippers;                                                                       virtual;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);            virtual;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);              virtual;
    procedure   GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler);            virtual;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);              virtual;
    procedure   AcceptParam( const aPath: string; aValue: TSignal);
    function    GetProfileInfoHtml: string;
    function    IsSameModule( const aModule: TMod): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                            virtual;
    procedure   SendMidi( const aMsg: TMidiMessage);                                              virtual;
    procedure   SampleRateChanged;                                                                virtual;
  public
    property    Name                             : string   read FName                  write SetName;
    property    ResetFlag                        : Boolean  read GetResetFlag           write SetResetFlag;
    property    SamplesFast                      : Int64    read FSamplesFast;
    property    SamplesSlow                      : Int64    read FSamplesSlow;
    property    InputCount                       : Integer  read GetInputCount;
    property    OutputCount                      : Integer  read GetOutputCount;
    property    Output[ anIndex: Integer]        : TSignal  read GetOutput;
    {$IFDEF PATCH_PROFILER}
  public
    property    TotalClocksFast                  : Int64    read FTotalClocksFast;
    property    TotalClocksSlow                  : Int64    read FTotalClocksSlow;
    property    TotalZips                        : Int64    read FTotalZips;
    {$ENDIF}
  public
    property    InternalValue      [ const aName: string] : TSignal                     write SetInternalValue;
    property    InternalStringValue[ const aName: string] : string                      write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string  read GetStringValue write SetStringValue;
  end;

  TModInput = class( TMod)
  strict private
  const
    o_out1 = 0;
    o_out2 = 1;
    o_out3 = 2;
    o_out4 = 3;
    o_out5 = 4;
    o_out6 = 5;
    o_out7 = 6;
    o_out8 = 7;
  public
    procedure   DoTick;                                                                          override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModOutput = class( TMod)
  strict private
  const
    i_in1 = 0;
    i_in2 = 1;
    i_in3 = 2;
    i_in4 = 3;
    i_in5 = 4;
    i_in6 = 5;
    i_in7 = 6;
    i_in8 = 7;
  public
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TSynthPatch = class( TMod)
  private
    FVersion          : Integer;
    FModules          : TModuleArray;
    FConnections      : TConnections;
    FDecimatorCount   : Integer;
    FInputFifo1       : TSignalPairFifo;
    FInputFifo2       : TSignalPairFifo;
    FInputFifo3       : TSignalPairFifo;
    FInputFifo4       : TSignalPairFifo;
    FOnSendMidi       : TOnSendMidi;
    FOnSynthLog       : TOnSynthLog;
    FOnSynthError     : TOnSynthError;
    {$IFDEF PATCH_PROFILER}
    FWireClocks       : Int64;
    FTotalWireClocks  : Int64;
    {$ENDIF}
  private
    function    AssertUniqueName( const aName: string): string;                                  override;
  private
    function    CheckModuleIndex( anIndex: Integer): Boolean;
    function    GetModuleCount: Integer;
    function    GetModule( anIndex: Integer): TMod;
    function    GetConnectionCount: Integer;
    procedure   SetResetFlag( aValue: Boolean);                                                  override;
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
  protected
    procedure   HandleParam( const aPath: string; aValue: TSignal);                              override;
    procedure   Log( const aMsg: string);                                                        override;
    procedure   Error( const aMsg: string);                                                      override;
    {$IFDEF PATCH_PROFILER}
  protected
    procedure   PreWires;
    procedure   PostWires;
    {$ENDIF}
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SaveConnections   ( anIndent: Integer; var S: string);
    procedure   SaveModules       ( anIndent: Integer; var S: string);
    procedure   SaveDetails       ( anIndent: Integer; var S: string);                           override;
    procedure   Clear;                                                                           override;
    procedure   ResetSampleCounts;                                                               override;
    procedure   RemoveConnectionsFor( aModuleIndex: Integer);
    procedure   ClearModules;
    function    FindModule( const aModuleName: string): Integer;
    function    ModuleName( anIndex: Integer): string;
    function    FindInputPin ( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    InputPinName ( aModuleIndex, aPinIndex: Integer): string;
    function    OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    procedure   AddModule( const aModule: TMod);
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);               overload;
    procedure   AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string);          overload;
    procedure   PropagateSignals;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    procedure   DoTick;                                                                          override;
    procedure   FixZippers;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    procedure   GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler);           override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    procedure   AcceptSamplePair1( aLeft, aRight: TSignal);
    procedure   AcceptSamplePair2( aLeft, aRight: TSignal);
    procedure   AcceptSamplePair3( aLeft, aRight: TSignal);
    procedure   AcceptSamplePair4( aLeft, aRight: TSignal);
    function    GetProfileHtml: TStringList;
    function    IsSamePatch       ( const aPatch: TSynthPatch): Boolean;
    function    SameModules       ( const aPatch: TSynthPatch): Boolean;
    function    SameConnections   ( const aPatch: TSynthPatch): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                           override;
    procedure   SendMidi( const aMsg: TMidiMessage);                                             override;
    procedure   SampleRateChanged;                                                               override;
    procedure   ClearOverUnderFlows;
  public
    property    Version                      : Integer       read FVersion       write FVersion;
    property    ModuleCount                  : Integer       read GetModuleCount;
    property    Module    [ anIndex: Integer]: TMod          read GetModule;
    property    ConnectionCount              : Integer       read GetConnectionCount;
    property    Overflows                    : Integer       read GetOverflows;
    property    Underflows                   : Integer       read GetUnderflows;
    property    OnSendMidi                   : TOnSendMidi   read FOnSendMidi    write FOnSendMidi;
    property    OnSynthLog                   : TOnSynthLog   read FOnSynthLog    write FOnSynthLog;
    property    OnSynthError                 : TOnSynthError read FOnSynthError  write FOnSynthError;
  end;

  procedure RegisterModuleType( aModuleId: Integer; aModuleClass: TModuleClass);
  function  LookupModuleClass( aModuleId: Integer): TModuleClass;

  procedure ParsePrefix ( const aPath: string; var aPrefix, aRest: string);
  function  ParsePostfix( const aPath: string): string;

  function  ParseTextSequencerValue( const aValue: string): TSignals;

  function  MakeMidiNoteOff        ( aVelocity, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiNoteOn         ( aVelocity, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiKeyPressure    ( aPressure, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiControlChange  ( aValue, aController, aChannel: Byte) : TMidiMessage;
  function  MakeMidiProgramChange  ( aProgram, aChannel: Byte)            : TMidiMessage;
  function  MakeMidiChannelPressure( aValue, aChannel: Byte)              : TMidiMessage;
  function  MakeMidiPitchBend      ( aData: Word; aChannel: Byte)         : TMidiMessage;
  function  MakeMidiSongPosition   ( aData: Word)                         : TMidiMessage;
  function  MakeMidiSongSelect     ( aSong: Byte)                         : TMidiMessage;
  function  MakeMidiTuneRequest                                           : TMidiMessage;
  function  MakeMidiTimingClock                                           : TMidiMessage;
  function  MakeMidiStart                                                 : TMidiMessage;
  function  MakeMidiContinue                                              : TMidiMessage;
  function  MakeMidiStop                                                  : TMidiMessage;
  function  MakeMidiActiveSense                                           : TMidiMessage;
  function  MakeMidiReset                                                 : TMidiMessage;
  function  MakeMidiSysExData      ( const aData: TBytes)                 : TMidiMessage;

  procedure MidiMessageToBytes( const aMsg: TMidiMessage; var aData: TBytes);
  function  MidiMessageToLog( const aMsg: TMidiMessage): string;
  function  SignalForInputType( const aSignal: TSignal; anInputType: Integer): TSignal;        inline;
  function  SignalForOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal;      inline;
  function  SignalForEnvOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal;   inline;
  function  PolyBLEP( aPhase, anAmount, aDeltaPhase: TSignal): TSignal;                        inline;



implementation



uses

  Globals;

var

  ModuleIdTypeMaps  : TModuleIdTypeMaps;


  procedure RegisterModuleType( aModuleId: Integer; aModuleClass: TModuleClass);
  var
    aClass : TModuleClass;
  begin
    Assert( Assigned( aModuleClass), 'Trying to register a NIL module type');
    aClass := LookupModuleClass( aModuleId);
    if not Assigned( aClass)
    then begin
      SetLength( ModuleIdTypeMaps, Length( ModuleIdTypeMaps) + 1);
      ModuleIdTypeMaps[ Length( ModuleIdTypeMaps) - 1].ModuleId    := aModuleId;
      ModuleIdTypeMaps[ Length( ModuleIdTypeMaps) - 1].ModuleClass := aModuleClass;
    end
    else raise EModuleDuplicate.CreateFmt( 'Duplicate module ID %d for already present type %s while trying to register type %s', [ aModuleId, aClass.ClassName, aModuleClass.ClassName]);
  end;

  function  LookupModuleClass( aModuleId: Integer): TModuleClass;
  var
    M : TModuleIdTypeMap;
  begin
    Result := nil;
    for M in ModuleIdTypeMaps
    do begin
      if M.ModuleId = aModuleId
      then begin
        Result := M.ModuleClass;
        Break;
      end;
    end;
  end;

  procedure ParsePrefix( const aPath: string; var aPrefix, aRest: string);
  // NOTE: this code needs be compiled with "String Format Checking" on,
  //       otherwise it will generate a run time error.
  //       ok .. this will only occur when aPath and aRest refer to the same
  //       underlying string .. as in ParsePrefix( aRest, aPrefix, aRest);
  var
    i  : Integer;
  begin
    i       := 1;
    aPrefix := '';
    aRest   := '';
    while i <= Length( aPath)
    do begin
      if aPath[ i] = '/'
      then begin
        aPrefix := Copy( aPath, 1, i - 1 );                    // Either on this line
        aRest := Copy( aPath, i + 1, Length( aPath));          // Or this one .. not quite clear.
        Break;
      end
      else Inc( i);
    end;
  end;

  function  ParsePostfix( const aPath: string): string;
  var
    i : Integer;
  begin
    Result := '';
    for i := Length( aPath) downto 1
    do begin
      if aPath[ i] = '_'
      then begin
        Result := Copy( aPath, i + 1, Length( aPath));
        Break;
      end
    end;
  end;

  function  ParseTextSequencerValue( const aValue: string): TSignals;
  var
    aStrings : TStringList;
    aString  : string;
    i        : Integer;
    j        : Integer;
    aSignal  : TSignal;
    aParts   : TStringList;
    aRepeats : Integer;
  begin
    Result   := TSignals.Create( 0);
    aStrings := Explode( Trim( aValue), ',');
    try
      for i := 0 to aStrings.Count - 1
      do begin
        aString := Trim( aStrings[ i]);
        aParts  := Explode( aString, ':');
        if aParts.Count > 0
        then begin
          if aParts.Count > 1
          then aRepeats := StrToIntDef( Trim( aParts[ 1]), 1)
          else aRepeats := 1;
          if ParseTextValue( Trim( aParts[ 0]), aSignal, System_Rate)
          then begin
            if IsNan( aSignal)
            then aSignal := NO_NOTE;
            Result.AddSignal( aSignal);
            for j := 1 to aRepeats - 1
            do Result.AddSignal( NO_NOTE);
          end
          else Result.AddSignal( NO_NOTE);
        end
        else Result.AddSignal( NO_NOTE);
      end;
    finally
      aStrings.Free;
    end;
  end;

  function  MakeMidiNoteOff( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOff;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiNoteOn( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOn;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiKeyPressure( aPressure, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CKeyPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aPressure;
  end;

  function  MakeMidiControlChange( aValue, aController, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CControlChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aController;
    Result.Data[ 1] := aValue;
  end;

  function  MakeMidiProgramChange( aProgram, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CProgramChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aProgram;
  end;

  function  MakeMidiChannelPressure( aValue, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CChannelPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aValue;
  end;

  function  MakeMidiPitchBend( aData: Word; aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CPitchBend;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongPosition( aData: Word): TMidiMessage;
  begin
    Result.Command := SSongPosition;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongSelect( aSong: Byte): TMidiMessage;
  begin
    Result.Command := SSongSelect;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aSong;
  end;

  function  MakeMidiTuneRequest: TMidiMessage;
  begin
    Result.Command := STuneRequest;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiTimingClock: TMidiMessage;
  begin
    Result.Command := STimingClock;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStart: TMidiMessage;
  begin
    Result.Command := SStart;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiContinue: TMidiMessage;
  begin
    Result.Command := SContinue;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStop: TMidiMessage;
  begin
    Result.Command := SStop;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiActiveSense: TMidiMessage;
  begin
    Result.Command := SActiveSense;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiReset: TMidiMessage;
  begin
    Result.Command := SReset;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiSysExData( const aData: TBytes): TMidiMessage;
  var
    i : Integer;
  begin
    Result.Command := SSysEx;
    Result.Channel := 0;
    SetLength( Result.Data, Length( aData));
    for i := 0 to Length( Result.Data) - 1
    do Result.Data[ i] := aData[ i];
  end;

  procedure MidiMessageToBytes( const aMsg: TMidiMessage; var aData: TBytes);
  begin
    SetLength( aData, 0);

    case aMsg.Command of

      CNoteOff,          // = $80; // Length = 3;
      CNoteOn,           // = $90; // Length = 3;
      CKeyPressure,      // = $a0; // Length = 3;
      CControlChange :   // = $b0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
            aData[ 2] := aMsg.Data[ 1];
          end;
        end;

      CProgramChange,    // = $c0; // Length = 2;
      CChannelPressure : // = $d0; // Length = 2;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 2);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
          end;
        end;

      CPitchBend :       // = $e0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command  and $00f0) or ( aMsg.Channel and $000f);
            aData[ 1] :=   aMsg.Data[ 0] and $007f;
            aData[ 2] := ( aMsg.Data[ 0] and $3f80) shr 7;
          end;
        end;

//      CSysEx :          // = $f0; // Length = n;
//        begin
//        end;

    end;
  end;

  function  MidiMessageToLog( const aMsg: TMidiMessage): string;

    function CmdToStr( aCmd: Byte): string;
    begin
      case aCmd of
        CNoteOff         : Result := 'note off';         // = $80; // Length = 3;
        CNoteOn          : Result := 'note on';          // = $90; // Length = 3;
        CKeyPressure     : Result := 'key pressure';     // = $a0; // Length = 3;
        CControlChange   : Result := 'control change';   // = $b0; // Length = 3;
        CProgramChange   : Result := 'program change';   // = $c0; // Length = 2;
        CChannelPressure : Result := 'channel pressure'; // = $d0; // Length = 2;
        CPitchBend       : Result := 'pitch bend';       // = $e0; // Length = 2;
        CSysEx           : Result := 'sysex';            // = $f0; // Length = n;
      end;
    end;

  var
    i : Integer;
  begin
    Result := Format( 'ch %d cmd %.2x [%s] ', [ aMsg.Channel + 1, aMsg.Command, CmdToStr( aMsg.Command)]);
    for i := 0 to Length( aMsg.Data) - 1
    do Result := Result + Format( '%.2x ', [ aMSg.Data[ i]]);
    SetLength( Result, Length( Result) - 1);
  end;

  function  SignalForInputType( const aSignal: TSignal; anInputType: Integer): TSignal; inline;
  // Translates input range to a [-1..1] output range
  begin
    case anInputType of
      inputtype_inverted : Result :=     - aSignal    ;
      inputtype_positive : Result :=   2 * aSignal - 1;
      inputtype_pos_inv  : Result := - 2 * aSignal + 1;
      inputtype_negative : Result :=   2 * aSignal + 1;
      inputtype_neg_inv  : Result := - 2 * aSignal - 1;
      else                 Result :=       aSignal    ;
    end;
  end;

  function  SignalForOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal; inline;
  // Translates a [-1..1] range to an output range
  begin
    case anOutputType of
      outputtype_inverted : Result :=       - aSignal       ;
      outputtype_positive : Result := (   1 + aSignal) * 0.5;
      outputtype_pos_inv  : Result := (   1 - aSignal) * 0.5;
      outputtype_negative : Result := ( - 1 + aSignal) * 0.5;
      outputtype_neg_inv  : Result := ( - 1 - aSignal) * 0.5;
      else                  Result :=         aSignal       ;
    end;
  end;

  function  SignalForEnvOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal; inline;
  // Translates a [-1..1] range to an output range
  begin
    case anOutputType of
      envoutputtype_inverted : Result :=   - aSignal    ;
      envoutputtype_negative : Result :=     aSignal - 1;
      envoutputtype_neg_inv  : Result := 1 - aSignal    ;
      else                     Result :=     aSignal    ;
    end;
  end;


  function  PolyBLEP( aPhase, anAmount, aDeltaPhase: TSignal): TSignal; inline;
  // Moved out of HrastUnit.pas for usage by other oscillators than HrastOsc as well.
  var
    dt: TSignal;
  begin
    dt := anAmount * aDeltaPhase;
    if aPhase < dt
    then begin
      aPhase := aPhase / dt - 1;
      Result := - aPhase * aPhase;
    end
    else if aPhase > 1 - dt
    then begin
      aPhase := ( aPhase - 1) / dt + 1;
      Result := aPhase * aPhase;
    end
    else Result := 0
  end;


{ ========
  TConnection = class
  private
    Peer      : TSynthPatch;
    Index     : Integer;
    SrcModule : Integer;
    SrcPin    : Integer;
    DstModule : Integer;
    DstPin    : Integer;
  private
}

    procedure   TConnection.Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
    begin
      Peer      := aPeer;
      Index     := anIndex;
      SrcModule := aSrcModule;
      SrcPin    := aSrcPin;
      DstModule := aDstModule;
      DstPin    := aDstPin;
    end;

//  public

    procedure   TConnection.Save( anIndent: Integer; var S: string);
    var
      aSrcModName : string;
      aSrcPinName : string;
      aDstModName : string;
      aDstPinName : string;
      T           : string;
    begin
      if Assigned( Peer)
      then begin
        aSrcModName := Peer.ModuleName   ( SrcModule);
        aSrcPinName := Peer.OutputPinName( SrcModule, SrcPin);
        aDstModName := Peer.ModuleName   ( DstModule);
        aDstPinName := Peer.InputPinName ( DstModule, DstPin);

        T := Format( '%sconnect( %s %s %s %s)' + sLineBreak, [ Indent( anIndent), aSrcModName, aSrcPinName, aDstModName, aDstPinName], AppLocale);
        S := S + T
      end;
    end;

    function    TConnection.IsSameConnection( aConnection: TConnection): Boolean;
    begin
      Result := False;
      if Assigned( aConnection)
      then
        Result :=
          ( SrcModule = aConnection.SrcModule) and
          ( SrcPin    = aConnection.SrcPin   ) and
          ( DstModule = aConnection.DstModule) and
          ( DstPin    = aConnection.DstPin   );
    end;


{ ========
  TConnections = class
  private
    FPeer       : TSynthPatch;
    FConnections: TConnectionArray;
  public
    property    Count: Integer                             read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  private
}

    function    TConnections.GetCount: Integer;
    begin
      Result := Length( FConnections);
    end;

    function    TConnections.GetConnection( anIndex: Integer): TConnection;
    begin
      Result := FConnections[ anIndex]
    end;

    procedure   TConnections.SetConnection( anIndex: Integer; const aValue: TConnection);
    begin
      FConnections[ anIndex] := aValue;
    end;

//  public

    constructor TConnections.Create( aPeer: TSynthPatch);
    begin
      inherited Create;
      FPeer := aPeer;
    end;

    destructor  TConnections.Destroy; // override;
    begin
      Clear;
      inherited;
    end;

    procedure   TConnections.Save( anIndent: Integer; const aName: string; var S: string);
    var
      i : Integer;
    begin
      S := S + Indent( anIndent) + aName + '(' + sLineBreak;
      for i := 0 to Count - 1
      do Connection[ i].Save( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TConnections.Clear;
    begin
      while Count > 0
      do RemoveConnection( Count - 1);
    end;

    procedure   TConnections.AddConnection( const aValue: TConnection); // overload;
    begin
      SetLength( FConnections, Count + 1);
      FConnections[ Count - 1] := aValue;
    end;

    procedure   TConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);  // overload;
    var
      aConnection: TConnection;
    begin
      aConnection := TConnection.Create;
      aConnection.Connect( FPeer, Count, aSrcModule, aSrcPin, aDstModule, aDstPin);
      AddConnection( aConnection);
    end;

    procedure   TConnections.RemoveConnection( anIndex: Integer); // overload;
    var
      aConnection: TConnection;
      i          : Integer;
    begin
      aConnection := Connection[ anIndex];
      if Assigned( aConnection)
      then begin
        for i := anIndex to Count - 2
        do Connection[ i] := Connection[ i + 1];
        SetLength( FConnections, Count - 1);
        aConnection.Free;
      end;
    end;

    procedure   TConnections.RemoveConnectionsFor( aModuleIndex: Integer);
    var
      i : Integer;
      C : TConnection;
    begin
      for i := Count - 1 downto 0
      do begin
        C := Connection[ i];
        if ( C.SrcModule = aModuleIndex) or ( C.DstModule = aModuleIndex)
        then RemoveConnection( i);
      end;
    end;

    procedure   TConnections.PropagateSignals( const aModules: TModuleArray);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do begin
        with Connection[ i]
        do aModules[ DstModule].FInputs[ DstPin] := aModules[ SrcModule].FOutputs[ SrcPin];
      end;
    end;

    function    TConnections.IsSameConnections( const aConnections: TConnections): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      if Assigned( aConnections) and ( Count = aConnections.Count)
      then begin
        Result := True;
        for i := 0 to Count - 1
        do begin
          if not Connection[ i].IsSameConnection( aConnections.Connection[ i])
          then begin
            Result := False;
            Break;
          end;
        end;
      end;
    end;


(* ========
  TMod = class
  private
    FParent          : TSynthPatch;   // The patch this module is part off
    FName            : string;        // A unique name for this module - unique within it's parent
    FResetFlag       : Integer;       // Set when a module reset was requested
    FSamplesFast     : Int64;         // Sample counter, incremented each time a new Fast sample is calcultaed
    FSamplesSlow     : Int64;         // Sample counter, incremented each time a new Slow sample is calcultaed
    {$IFDEF PATCH_PROFILER}
  private
    FClocksFast      : Int64;         // CPU Fast clocks for current cycle
    FClocksSlow      : Int64;         // CPU Slow clocks for current cycle
    FTotalClocksFast : Int64;         // CPU Fast clocks accumulated over the module life time
    FTotalClocksSlow : Int64;         // CPU Slow clocks accumulated over the module life time
    FZips            : Int64;         // CPU clocks spent on dezoppering for the current cycle
    FTotalZips       : Int64;         // CPU clocks acumulated FZips over the module life time
    {$ENDIF}
  protected
    FIsFast          : Boolean;       // True when fast mode processing must be performed
    FIsSlow          : Boolean;       // True when slow mode processing must be performed
    FIsMidi          : Boolean;       // True when midi reception must be enabled
    FInputs          : TSignalArray;  // The input signals, these include both patch signals and user controls
    FDezips          : TSignalArray;  // An extra set of inputs for user controls, these are copied to FInCache with some filtering when the signal happens to be be in the deziper map
    FOutputs         : TSignalArray;  // The ouput signals.
    FDezipperMap     : TKnobMap;      // Determines  what signals should be dezippered, most, but not all, knobs should be dezippered
  public
    property    Name                             : string   read FName                  write SetName;
    property    ResetFlag                        : Boolean  read GetResetFlag           write SetResetFlag;
    property    SamplesFast                      : Int64    read FSamplesFast;
    property    SamplesSlow                      : Int64    read FSamplesSlow;
    property    InputCount                       : Integer  read GetInputCount;
    property    OutputCount                      : Integer  read GetOutputCount;
    property    Output[ anIndex: Integer]        : TSignal  read GetOutput;
    {$IFDEF PATCH_PROFILER}
  public
    property    TotalClocksFast                  : Int64    read FTotalClocksFast;
    property    TotalClocksSlow                  : Int64    read FTotalClocksSlow;
    property    TotalZips                        : Int64    read FTotalZips;
    {$ENDIF}
  public
    property    InternalValue      [ const aName: string] : TSignal                     write SetInternalValue;
    property    InternalStringValue[ const aName: string] : string                      write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string  read GetStringValue write SetStringValue;
  private
*)

    function    TMod.AssertUniqueName( const aName: string): string; // virtual;
    begin
      if Assigned( FParent)
      then Result := FParent.AssertUniqueName( aName)
      else Result := aName;
    end;

//  private

    procedure   TMod.SetName( const aValue: string);
    begin
      if aValue <> FName
      then FName := AssertUniqueName( aValue);
    end;

    function    TMod.GetResetFlag: Boolean; // virtual;
    begin
      Result := FResetFlag > 0;
    end;

    procedure   TMod.SetResetFlag( aValue: Boolean); // virtual;
    begin
      if aValue
      then FResetFlag := 1
      else begin
        if FResetFlag > 0
        then Dec( FResetFlag);
      end;
    end;

    function    TMod.GetInputCount: Integer;
    begin
      Result := Length( FInputs);
    end;

    function    TMod.GetOutputCount: Integer;
    begin
      Result := Length( FOutputs);
    end;

    function    TMod.GetOutput( anIndex: Integer): TSignal;
    begin
      if ( anIndex >= 0) and ( anIndex < OutputCount)
      then Result := FOutputs[ anIndex]
      else Result := 0.0;
    end;

    procedure   TMod.SetCommonValue( anIndex: Integer; aValue: TSignal); // virtual;
    begin
      if anIndex >= 0
      then begin
        if anIndex in FDezipperMap
        then FDezips[ anIndex] := aValue
        else FInputs[ anIndex] := aValue;
      end;
    end;

    procedure   TMod.SetInternalValue( const aName: string; aValue: TSignal); // virtual;
    var
      anIndex: Integer;
    begin
      anIndex := FindInput( aName);
      if anIndex >= 0
      then SetCommonValue( anIndex, aValue)
      else SetInternal( aName, aValue);
    end;

    procedure   TMod.SetInternalStringValue( const aName, aValue: string);
    var
      aSignal : TSignal;
    begin
      if ParseTextValue( aValue, aSignal, System_Rate)
      then InternalValue[ aName] := aSignal
      else raise EUnitsUndefined.CreateFmt( 'Unparsable value ''%s''', [ aName])
    end;

//  protected

    procedure   TMod.SetStringValue( const aName, aValue: string); // virtual;
    begin
      // Nothing here.
    end;

    function    TMod.GetStringValue( const aName: string): string; // virtual;
    begin
      Result := '';
    end;

    procedure   TMod.FixInputs;
    var
      aCount : Integer;
    begin
      aCount := 0;
      while not SameText( InputName( aCount), NO_INPUT)
      do begin
        Inc( aCount);
        if aCount > MAX_INPUTS
        then raise EInputNotFound.CreateFmt( 'are there really over %d inputs? (probably %s.InputName is broken and never returns NO_INPUT)', [ MAX_INPUTS, ClassName]);
      end;
      SetLength( FInputs, aCount);
      SetLength( FDezips, aCount);
    end;

    procedure   TMod.FixOutputs;
    var
      aCount : Integer;
    begin
      aCount := 0;
      while not SameText( OutputName( aCount), NO_OUTPUT)
      do begin
        Inc( aCount);
        if aCount > MAX_OUTPUTS
        then raise EOutputNotFound.CreateFmt( 'are there really over %d outputs? (probably %s.OutputName is broken and never returns NO_OUTPUT)', [ MAX_OUTPUTS, ClassName]);
      end;
      SetLength( FOutputs, aCount);
    end;

    procedure   TMod.SetInputs( aStartIndex: Integer; const aValue: TSignals);
    var
      i : Integer;
    begin
      SetLength( FInputs, aStartIndex + aValue.Count);
      SetLength( FDezips, aStartIndex + aValue.Count);
      for i := aStartIndex to Length( FInputs) - 1
      do begin
        FInputs[ i] := aValue.Signal[ i - aStartIndex];
        FDezips[ i] := aValue.Signal[ i - aStartIndex];
      end;
    end;

    procedure   TMod.HandleParam( const aControl: string; aValue: TSignal); // virtual;
    // Handle a parameter set from the editor patch
    begin
      InternalValue[ aControl] := aValue;
    end;

    procedure   TMod.Log( const aMsg: string); // virtual;
    begin
      if Assigned( FParent)
      then FParent.Log( aMsg);
    end;

    procedure   TMod.LogFmt( const aFmt: string; const anArgs: array of const);
    begin
      Log( Format( aFmt, anArgs, AppLocale));
    end;

    procedure   TMod.Error( const aMsg: string); // virtual;
    begin
      if Assigned( FParent)
      then FParent.Error( aMsg);
    end;

    procedure   TMod.ErrorFmt( const aFmt: string; const anArgs: array of const);
    begin
      Error( Format( aFmt, anArgs, AppLocale));
    end;

    {$IFDEF PATCH_PROFILER}

//  protected

    procedure   TMod.PreTickFast;
    begin
      {$Q-R-}
      FClocksFast := GetCpuClockCycleCount;
      {$Q+R+}
    end;

    procedure   TMod.PreTickSlow;
    begin
      {$Q-R-}
      FClocksSlow := GetCpuClockCycleCount;
      {$Q+R+}
    end;

    procedure   TMod.PostTickFast;
    begin
      {$Q-R-}
      FClocksFast      := GetCpuClockCycleCount - FClocksFast;
      FTotalClocksFast := FTotalClocksFast      + FClocksFast;
      {$Q+R+}
    end;

    procedure   TMod.PostTickSlow;
    begin
      {$Q-R-}
      FClocksSlow      := GetCpuClockCycleCount - FClocksSlow;
      FTotalClocksSlow := FTotalClocksSlow + FClocksSlow;
      {$Q+R+}
    end;

    procedure   TMod.PreZipper;
    begin
      {$Q-R-}
      FZips := GetCpuClockCycleCount;
      {$Q+R+}
    end;

    procedure   TMod.PostZipper;
    begin
      {$Q-R-}
      FZips      := GetCpuClockCycleCount - FZips;
      FTotalZips := FTotalZips + FZips;
      {$Q+R+}
    end;

    {$ENDIF}

//  public

    constructor TMod.Create( aParent: TSynthPatch; const aName: string); // virtual;
    begin
      inherited Create;
      {$IFDEF PATCH_PROFILER}
      FTotalClocksSlow := 0;
      FTotalClocksFast := 0;
      FTotalZips       := 0;
      {$ENDIF}
      FIsFast          := False;
      FIsSlow          := False;
      FIsMidi          := False;
      FParent          := aParent;
      Name             := aName;
      FixOutputs;
      FixInputs;
      SetDefaults;
    end;

    destructor  TMod.Destroy; // override;
    begin
      FParent := nil;
      inherited;
    end;

    procedure   TMod.SetDefaults; // virtual;
    begin
    end;

    procedure   TMod.SaveKeyValue( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    var
      aBaseName : string;
      i         : Integer;
      Output    : string;
    begin
      aBaseName := '';
      for i := 1 to Length( aName)
      do begin
        if CharInSet( aName[ i], IdentifierStarters)
        then abaseName := abaseName + aName[ i]
        else Break;
      end;
      Output := Format( '%s', [ SignalToStr( aValue)], AppLocale);
      S := S + Format( Indent( anIndent + 1) + '%s %s' + sLineBreak, [ aName, Output], AppLocale);
    end;

    procedure   TMod.SaveInputValues( anIndent: Integer; var S: string);
    var
      i : Integer;
      N : string;
      p : Integer;
    begin
      p := 1;
      S := S + Indent( anIndent) + 'inputs(' + sLineBreak;
      for i := 0 to InputCount - 1
      do begin
        N := InputName( i);
        if SameText( N, NO_INPUT)
        then begin
          N := Format( 'in%d', [ p], AppLocale);
          Inc( p);
        end;
        SaveKeyValue( anIndent, N, FInputs[ i], S);
      end;
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TMod.SaveDetails( anIndent: Integer; var S: string); // virtual;
    begin
      S := S + Format( Indent( anIndent) + '%s' + sLineBreak, [ ClassName ], AppLocale);
      S := S + Format( Indent( anIndent) + '%s' + sLineBreak, [ Name      ], AppLocale);
      SaveInternals  ( anIndent, S);
      SaveInputValues( anIndent, S);
    end;

    procedure   TMod.SaveInternals( anIndent: Integer; var S: string); // virtual;
    var
      aName: string;
      i    : Integer;
    begin
      aName := InternalName( 0);
      if not SameText( aName, NO_INTERNAL)
      then begin
        S := S + Indent( anIndent) + 'internals(' + sLineBreak;
        i := 0;
        while not SameText( aName, NO_INTERNAL)
        do begin
          SaveKeyValue( anIndent, aName, GetInternal( aName), S);
          Inc( i);
          if i > MAX_INTERNALS
          then raise EInternalNotFound.CreateFmt( 'are there really over %d internals? (probably %sInternalName is broken and never returns NO_INTERNAL)', [ MAX_INTERNALS, ClassName]);
          aName := InternalName( i);
        end;
        S := S + Indent( anIndent) + ')' + sLineBreak;
      end;
    end;

    procedure   TMod.SetInternal( const aName: string; aValue: TSignal); // virtual;
    begin
      // Internals is an option not present here.
    end;

    function    TMod.GetInternal( const aName: string): TSignal; // virtual;
    begin
      // Internals is an option not present here.
      Result := UNDEFINED;
    end;

    procedure   TMod.Save( anIndent: Integer; var S: string);
    begin
      S := S + Indent( anIndent) + 'module(' + sLineBreak;
      SaveDetails  ( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TMod.Reset; // virtual;
    begin
      ResetFlag := True;
    end;

    procedure   TMod.ResetSampleCounts;
    begin
      FSamplesFast := 0;
      FSamplesSlow := 0;
    end;

    procedure   TMod.Clear; // virtual;
    begin
    end;

    function    TMod.FindInput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;
      for i := 0 to InputCount - 1
      do begin
        if SameText( aPinName, InputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TMod.FindOutput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;
      for i := 0 to OutputCount - 1
      do begin
        if SameText( aPinName, OutputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TMod.InputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_INPUT;
    end;

    function    TMod.InternalName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_INTERNAL;
    end;

    function    TMod.OutputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_OUTPUT;
    end;

    procedure   TMod.Tick;
    begin
      {$Q-R-}
      Inc( FSamplesFast);
      {$Q+R+}
      {$IFDEF PATCH_PROFILER}PreTickFast;{$ENDIF}
      DoTick;
      {$IFDEF PATCH_PROFILER}PostTickFast;{$ENDIF}
    end;

    procedure   TMod.SlowTick;
    begin
      {$Q-R-}
      Inc( FSamplesSlow);
      {$Q+R+}
      {$IFDEF PATCH_PROFILER}PreTickSlow;{$ENDIF}
      DoSlowTick;
      {$IFDEF PATCH_PROFILER}PostTickSlow;{$ENDIF}
    end;

    procedure   TMod.DeZipper;
    const
      Speed = 1.0 / 512.0;
    var
      i : Integer;
    begin
      {$IFDEF PATCH_PROFILER}PreZipper;{$ENDIF}
      for i := 0 to Length( FInputs) - 1
      do begin
        if i in FDezipperMap
        then FInputs[ i] := Normalize( Speed * ( FDezips[ i] - FInputs[ i]) + FInputs[ i]);
      end;
      {$IFDEF PATCH_PROFILER}PostZipper;{$ENDIF}
    end;

    procedure   TMod.DoTick; // virtual;
    begin
      // audio rate
    end;

    procedure   TMod.DoSlowTick; // virtual;
    begin
      // control rate
    end;

    procedure   TMod.FixZippers; // virtual;
    var
      i : Integer;
    begin
      for i := 0 to Length( FInputs) - 1
      do begin
        if i in FDezipperMap
        then FInputs[ i] := FDezips[ i];
      end;
    end;

    procedure   TMod.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // virtual;
    begin
      // Collect info to be sent to module lights - this module has no lights
    end;

    procedure   TMod.GatherData( const aPrefix: string; const aCallback: TDataHandler); // virtual;
    begin
      // Collect numerical info to be sent to module viewers - this module has no viewers
    end;

    procedure   TMod.GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler); // virtual;
    begin
      // Collect numerical info to be sent to module XY viewers - this module has no XY viewers
    end;

    procedure   TMod.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // virtual;
    begin
      // Collect string info to be sent to module viewers - this module has no viewers
    end;

    procedure   TMod.AcceptParam( const aPath: string; aValue: TSignal);
    // Accept knob data sent from visual modules
    var
      aPrefix : string;
      aParam  : string;
      p       : Integer;
    begin
      if ( ClassType = TSynthPatch)
      then HandleParam( aPath, aValue)
      else begin
        ParsePrefix( aPath, aPrefix, aParam);
        if SameText( aPrefix, Name)                      // is the message for us?
        then begin
          if Pos( '/', aParam) = 0                       // When no more slashes left
          then begin                                     // we must be at the final bit ...
            p := Pos( '_', aParam);                      // The xxx_ is a conveniance in the "desiger" ... strip it here
            if p > 0
            then begin
              aParam := Copy( aParam, p + 1, Length( aParam));
              HandleParam( aParam, aValue);
            end;
          end;
        end;
      end;
    end;

    function    TMod.GetProfileInfoHtml: string;
    {$IFDEF PATCH_PROFILER}
    var
      FR : string;
      SR : string;
    {$ENDIF}
    begin
    {$IFDEF PATCH_PROFILER}
      if SamplesFast = 0
      then FR := '---'
      else FR := Format( '%.2f', [ TotalClocksFast / SamplesFast], AppLocale);
      if SamplesSlow = 0
      then SR := '---'
      else SR := Format( '%.2f', [ TotalClocksSlow / SamplesSlow / Control_Decimation], AppLocale);

      Result :=
        Format(
          '<tr><td>%s</td> <td>%s</td> <td align="right">%d</td> <td align="right">%d</td> <td align="right">%d</td> <td align="right">%d</td> <td align="right">%s</td> <td align="right">%s</td></tr>',
          [
            Name,
            ClassName,
            SamplesFast,
            SamplesSlow,
            TotalClocksFast,
            TotalClocksSlow,
            FR,
            SR
          ],
          AppLocale
        );
      {$ELSE}
        Result := '';
      {$ENDIF}
    end;

    function    TMod.IsSameModule( const aModule: TMod): Boolean;
    begin
      Result := False;
      if Assigned( aModule)
      then Result := ClassType = aModule.ClassType;
    end;

    procedure   TMod.AcceptMidi( const aMsg: TMidiMessage); // virtual;
    begin
      // Nothing here, if is for chuldren, just building daydreams ...
    end;

    procedure   TMod.SendMidi( const aMsg: TMidiMessage);
    begin
      if Assigned( FParent)
      then FParent.SendMidi( aMsg);
    end;

    procedure   TMod.SampleRateChanged; // virtual;
    // Notification for sample rate change, the actual values are in global variables
    // System_Rate for the audio rate and Control_Rate for the control rate.
    // Modules caching sample rate dependent data should recalculate such data in an
    // overriden version of this procedure.
    begin
      // Nothing for abstract modules.
    end;


{ ========
  TModInput = class( TMod)
  public
}

    procedure   TModInput.DoTick; // override;
    begin
    end;

    function    TModInput.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        o_out5 : Result := 'out5';
        o_out6 : Result := 'out6';
        o_out7 : Result := 'out7';
        o_out8 : Result := 'out8';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModOutput = class( TMod)
  public
}

    function    TModOutput.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1 : Result := 'in1';
        i_in2 : Result := 'in2';
        i_in3 : Result := 'in3';
        i_in4 : Result := 'in4';
        i_in5 : Result := 'in5';
        i_in6 : Result := 'in6';
        i_in7 : Result := 'in7';
        i_in8 : Result := 'in8';
        else    Result := NO_INPUT;
      end;
    end;


(* ========
  TSynthPatch = class
  private
    FVersion          : Integer;
    FModules          : TModuleArray;
    FConnections      : TConnections;
    FDecimatorCount   : Integer;
    FInputFifo1       : TSignalPairFifo;
    FInputFifo2       : TSignalPairFifo;
    FInputFifo3       : TSignalPairFifo;
    FInputFifo4       : TSignalPairFifo;
    FOnSendMidi       : TOnSendMidi;
    FOnSynthLog       : TOnSynthLog;
    FOnSynthError     : TOnSynthError;
    {$IFDEF PATCH_PROFILER}
    FWireClocks       : Int64;
    FTotalWireClocks  : Int64;
    {$ENDIF}
  public
    property    Version                      : Integer       read FVersion       write FVersion;
    property    ModuleCount                  : Integer       read GetModuleCount;
    property    Module    [ anIndex: Integer]: TMod          read GetModule;
    property    ConnectionCount              : Integer       read GetConnectionCount;
    property    Overflows                    : Integer       read GetOverflows;
    property    Underflows                   : Integer       read GetUnderflows;
    property    OnSendMidi                   : TOnSendMidi   read FOnSendMidi    write FOnSendMidi;
    property    OnSynthLog                   : TOnSynthLog   read FOnSynthLog    write FOnSynthLog;
    property    OnSynthError                 : TOnSynthError read FOnSynthError  write FOnSynthError;
  private
*)

    function    TSynthPatch.AssertUniqueName( const aName: string): string; // override;
    var
      i : Integer;
    begin
      // calling inherited here would go recursive
      Result := aName;
      for i := 0 to ModuleCount - 1
      do begin
        if SameText( FModules[ i].Name, aName)
        then raise EModule.CreateFmt( 'module name "%s" is not unique.', [ aName]);
      end;
    end;

//  private

    function    TSynthPatch.CheckModuleIndex( anIndex: Integer): Boolean;
    begin
      Result := ( anIndex >= 0) and ( anIndex < ModuleCount);
    end;

    function    TSynthPatch.GetModuleCount: Integer;
    begin
      Result := Length( FModules);
    end;

    function    TSynthPatch.GetModule( anIndex: Integer): TMod;
    begin
      Result := nil;
      if CheckModuleIndex( anIndex)
      then Result := FModules[ anIndex];
    end;

    function    TSynthPatch.GetConnectionCount: Integer;
    begin
      Result := 0;
      if Assigned( FConnections)
      then Result := FConnections.Count;
    end;

    procedure   TSynthPatch.SetResetFlag( aValue: Boolean); // override;
    begin
      if aValue
      then FResetFlag := 2 * Control_Decimation + 1
      else begin
        if FResetFlag > 0
        then Dec( FResetFlag);
      end;
    end;

    function    TSynthPatch.GetOverflows: Integer;
    begin
      Result := 0;
      if Assigned( FInputFifo1)
      then Result := Result + FInputFifo1.Overflows;
      if Assigned( FInputFifo2)
      then Result := Result + FInputFifo2.Overflows;
      if Assigned( FInputFifo3)
      then Result := Result + FInputFifo3.Overflows;
      if Assigned( FInputFifo4)
      then Result := Result + FInputFifo4.Overflows;
    end;

    function    TSynthPatch.GetUnderflows: Integer;
    begin
      Result := 0;
      if Assigned( FInputFifo1)
      then Result := Result + FInputFifo1.Underflows;
      if Assigned( FInputFifo2)
      then Result := Result + FInputFifo2.Underflows;
      if Assigned( FInputFifo3)
      then Result := Result + FInputFifo3.Underflows;
      if Assigned( FInputFifo4)
      then Result := Result + FInputFifo4.Underflows;
    end;

//  protected

    procedure   TSynthPatch.HandleParam( const aPath: string; aValue: TSignal); // override;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].AcceptParam( aPath, aValue);
    end;

    procedure   TSynthPatch.Log( const aMsg: string); // override;
    begin
      if Assigned( FOnSynthLog)
      then FOnSynthLog( Self, aMsg);
    end;

    procedure   TSynthPatch.Error( const aMsg: string); // override;
    begin
      if Assigned( FOnSynthError)
      then FOnSynthError( Self, aMsg);
    end;

    {$IFDEF PATCH_PROFILER}

//  protected

    procedure   TSynthPatch.PreWires;
    begin
      {$Q-R-}
      FWireClocks := GetCpuClockCycleCount;
      {$Q+R+}
    end;

    procedure   TSynthPatch.PostWires;
    begin
      {$Q-R-}
      FWireClocks      := GetCpuClockCycleCount - FWireClocks;
      FTotalWireClocks := FTotalWireClocks      + FWireClocks;
      {$Q+R+}
    end;

    {$ENDIF}

//  public

    constructor TSynthPatch.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast      := True;
      FIsSlow      := False;
      FIsMidi      := True;
      FVersion     := 0;
      FInputFifo1  := TSignalPairFifo.Create( FIFO_SIZE);
      FInputFifo2  := TSignalPairFifo.Create( FIFO_SIZE);
      FInputFifo3  := TSignalPairFifo.Create( FIFO_SIZE);
      FInputFifo4  := TSignalPairFifo.Create( FIFO_SIZE);
      FConnections := TConnections.Create( Self );
    end;

    destructor  TSynthPatch.Destroy; // override;
    begin
      FreeAndNil( FConnections);  // Do this first for a faster clear
      Clear;                      // Clears the modules - that is it removes and frees them
      FreeAndNil( FInputFifo1);
      FreeAndNil( FInputFifo2);
      FreeAndNil( FInputFifo3);
      FreeAndNil( FInputFifo4);
      inherited;
    end;

    procedure   TSynthPatch.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].SetDefaults;
    end;

    procedure   TSynthPatch.SaveConnections( anIndent: Integer; var S: string);
    begin
      if Assigned( FConnections)
      then FConnections.Save( anIndent, 'connections', S);
    end;

    procedure   TSynthPatch.SaveModules( anIndent: Integer; var S: string);
    var
      i : Integer;
    begin
      S := S + Indent( anIndent) + 'modules(' + sLineBreak;
      for i := 0 to ModuleCount - 1
      do Module[ i].Save( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TSynthPatch.SaveDetails( anIndent: Integer; var S: string); // override;
    begin
      inherited;
      S := S + Indent( anIndent) + 'version ' + IntToStr( Version) + sLineBreak;
      SaveModules    ( anIndent, S);
      SaveConnections( anIndent, S);
    end;

    procedure   TSynthPatch.Clear;   // Clear moduleas and connections
    begin
      ClearModules;
      inherited Clear;
    end;

    procedure   TSynthPatch.ResetSampleCounts; // override;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].ResetSampleCounts;
      inherited;
    end;

    procedure   TSynthPatch.RemoveConnectionsFor( aModuleIndex: Integer);
    begin
      if Assigned( FConnections)
      then FConnections.RemoveConnectionsFor( aModuleIndex);
    end;

    procedure   TSynthPatch.ClearModules;
    var
      aModule : TMod;
    begin
      while ModuleCount > 0
      do begin
        aModule :=  FModules[ ModuleCount - 1];
        RemoveConnectionsFor( ModuleCount - 1);
        SetLength(  FModules, ModuleCount - 1);
        aModule.Free;
      end;
    end;

    function    TSynthPatch.FindModule( const aModuleName: string): Integer;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;
      for i := 0 to ModuleCount - 1
      do begin
        if SameText( aModuleName, Module[ i].Name)
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TSynthPatch.ModuleName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ModuleCount)
      then Result := Module[ anIndex].Name
      else Result := NO_MODULE;
    end;

    function    TSynthPatch.FindInputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_PMC;
      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindInput( aPinName);
    end;

    function    TSynthPatch.FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_PMC;
      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindOutput( aPinName);
    end;

    function    TSynthPatch.InputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_INPUT;
      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].InputName( aPinIndex);
    end;

    function    TSynthPatch.OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_OUTPUT;
      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].OutputName( aPinIndex);
    end;

    procedure   TSynthPatch.AddModule( const aModule: TMod); // overload;
    begin
      SetLength( FModules, ModuleCount + 1);
      FModules[ ModuleCount - 1] := aModule;
    end;

    procedure   TSynthPatch.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer); // overload;
    begin
      if Assigned( FConnections)
      then FConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin);
    end;

    procedure   TSynthPatch.AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string); // overload;
    var
      aSrcM: Integer;
      aSrcP: Integer;
      aDstM: Integer;
      aDstP: Integer;
    begin
      aSrcM := FindModule( aSrcModule);
      if aSrcM >= 0
      then begin
        aSrcP := FindOutputPin( aSrcM, aSrcPin);
        if aSrcP >= 0
        then begin
          aDstM := FindModule( aDstModule);
          if aDstM >= 0
          then begin
            aDstP := FindInputPin( aDstM, aDstPin);
            if aDstP >= 0
            then AddConnection( aSrcM, aSrcP, aDstM, aDstP)
            else raise EInputNotFound.CreateFmt( 'input "%s" not found on destination module "%s"', [ aDstPin, aDstModule]);
          end
          else raise EModuleNotFound.CreateFmt( 'destination module "%s" not found', [ aDstModule]);
        end
        else raise EOutputNotFound.CreateFmt( 'output "%s" not found on source module "%s"', [ aSrcPin, aSrcModule]);
      end
      else raise EModuleNotFound.CreateFmt( 'source module "%s" not found', [ aSrcModule]);
    end;

    procedure   TSynthPatch.PropagateSignals;
    // Copy signals from inputs to outputs
    begin
      {$IFDEF PATCH_PROFILER}PreWires;{$ENDIF}
      FConnections.PropagateSignals( FModules);
      {$IFDEF PATCH_PROFILER}PostWires;{$ENDIF}
    end;

const

  synthpatch_in1  = 0;
  synthpatch_in2  = 1;
  synthpatch_in3  = 2;
  synthpatch_in4  = 3;
  synthpatch_in5  = 4;
  synthpatch_in6  = 5;
  synthpatch_in7  = 6;
  synthpatch_in8  = 7;

  synthpatch_out1 = 0;
  synthpatch_out2 = 1;
  synthpatch_out3 = 2;
  synthpatch_out4 = 3;
  synthpatch_out5 = 4;
  synthpatch_out6 = 5;
  synthpatch_out7 = 6;
  synthpatch_out8 = 7;

    function    TSynthPatch.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        synthpatch_in1 : Result := 'in1';
        synthpatch_in2 : Result := 'in2';
        synthpatch_in3 : Result := 'in3';
        synthpatch_in4 : Result := 'in4';
        synthpatch_in5 : Result := 'in5';
        synthpatch_in6 : Result := 'in6';
        synthpatch_in7 : Result := 'in7';
        synthpatch_in8 : Result := 'in8';
        else             Result := NO_INPUT;
      end;
    end;

    function    TSynthPatch.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        synthpatch_out1 : Result := 'out1';
        synthpatch_out2 : Result := 'out2';
        synthpatch_out3 : Result := 'out3';
        synthpatch_out4 : Result := 'out4';
        synthpatch_out5 : Result := 'out5';
        synthpatch_out6 : Result := 'out6';
        synthpatch_out7 : Result := 'out7';
        synthpatch_out8 : Result := 'out8';
        else              Result := NO_OUTPUT;
      end;
    end;

    procedure   TSynthPatch.DoTick; // Override;
    var
      i         : Integer;
      p         : Integer;
      aModCount : Integer;
      aMod      : TMod;
      aPair1    : TSignalPair;
      aPair2    : TSignalPair;
      aPair3    : TSignalPair;
      aPair4    : TSignalPair;
      Ch1       : TSignal;
      Ch2       : TSignal;
      Ch3       : TSignal;
      Ch4       : TSignal;
      Ch5       : TSignal;
      Ch6       : TSignal;
      Ch7       : TSignal;
      Ch8       : TSignal;
      OldEM     : TArithmeticExceptionMask;
    begin
      try
        // Get patch external inputs

        aPair1 := FInputFifo1.GetData;                                       // Fifo 1 always should have data
        if FInputFifo2.FillCount > 0 then aPair2 := FInputFifo2.GetData;
        if FInputFifo3.FillCount > 0 then aPair3 := FInputFifo3.GetData;
        if FInputFifo4.FillCount > 0 then aPair4 := FInputFifo4.GetData;

        aModCount       := ModuleCount;
        p               := 0;
        Ch1             := 0;
        Ch2             := 0;
        Ch3             := 0;
        Ch4             := 0;
        Ch5             := 0;
        Ch6             := 0;
        Ch7             := 0;
        Ch8             := 0;
        FDecimatorCount := ( FDecimatorCount + 1) mod Control_Decimation;

        {$IFDEF DEBUG_DENORMS}
        OldEM := SetExceptionMask( GetExceptionMask - [ exDenormalized]); // For debugging trap denormalized exceptions in addition to the usual ones
        {$ELSE}
        OldEM := SetExceptionMask( exAllArithmeticExceptions);            // For non-debug code disable all floating point exceptions
        {$ENDIF}

        try
          for i := 0 to aModCount - 1                                     // Tick the modules in the patch
          do begin
            aMod := Module[ i];
            with aMod
            do begin
              if Self.ResetFlag
              then aMod.Reset;

              if aMod is TModInput                                        // When module is an input module it needs the current in sample
              then begin
                FOutputs[ 0] := Normalize( aPair1.Left );
                FOutputs[ 1] := Normalize( aPair1.Right);
                FOutputs[ 2] := Normalize( aPair2.Left );
                FOutputs[ 3] := Normalize( aPair2.Right);
                FOutputs[ 4] := Normalize( aPair3.Left );
                FOutputs[ 5] := Normalize( aPair3.Right);
                FOutputs[ 6] := Normalize( aPair4.Left );
                FOutputs[ 7] := Normalize( aPair4.Right);
              end;

              if FIsFast
              then Tick;

              if p mod Control_Decimation = FDecimatorCount
              then begin
                if FIsSlow
                then SlowTick;
                DeZipper;
              end;
              Inc( p);

              if aMod is TModOutput                                       // When module is an output module add its inputs to our outputs
              then begin
                Ch1 := Ch1 + Normalize( FInputs[ 0]);
                Ch2 := Ch2 + Normalize( FInputs[ 1]);
                Ch3 := Ch3 + Normalize( FInputs[ 2]);
                Ch4 := Ch4 + Normalize( FInputs[ 3]);
                Ch5 := Ch5 + Normalize( FInputs[ 4]);
                Ch6 := Ch6 + Normalize( FInputs[ 5]);
                Ch7 := Ch7 + Normalize( FInputs[ 6]);
                Ch8 := Ch8 + Normalize( FInputs[ 7]);
              end;
            end;
          end;
          ResetFlag := False;
          PropagateSignals;                                              // Transfer module outputs to module inputs according to modulee wiring

          FOutputs[ 0] := Ch1;                                            // Finally fix our own outputs
          FOutputs[ 1] := Ch2;
          FOutputs[ 2] := Ch3;                                            // Finally fix our own outputs
          FOutputs[ 3] := Ch4;
          FOutputs[ 4] := Ch5;                                            // Finally fix our own outputs
          FOutputs[ 5] := Ch6;
          FOutputs[ 6] := Ch7;                                            // Finally fix our own outputs
          FOutputs[ 7] := Ch8;
        finally
          ClearFPUExceptions( False);
          SetExceptionMask( OldEM);
        end;
     except
        on E: Exception
        do KilledException( E);                                           // Always kill all exceptions here.
     end;
    end;

    procedure   TSynthPatch.FixZippers; // override;
    var
      i  : Integer;
    begin
      inherited;
      for i := 0 to ModuleCount - 1
      do Module[ i].FixZippers;
    end;

    procedure   TSynthPatch.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i  : Integer;
    begin
      try
        if aPrefix = ''
        then begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherLights( Name, aCallback);
        end
        else begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherLights( aPrefix + '/' + Name, aCallback);
        end;
      except
        on E: Exception
        do KilledException( E);
      end;
    end;

    procedure   TSynthPatch.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      i  : Integer;
    begin
      try
        if aPrefix = ''
        then begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherData( Name, aCallback);
        end
        else begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherData( aPrefix + '/' + Name, aCallback);
        end;
      except
        on E: Exception
        do KilledException( E);
      end;
    end;

    procedure   TSynthPatch.GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler); // override;
    var
      i  : Integer;
    begin
      try
        if aPrefix = ''
        then begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherXYData( Name, aCallback);
        end
        else begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherXYData( aPrefix + '/' + Name, aCallback);
        end;
      except
        on E: Exception
        do KilledException( E);
      end;
    end;

    procedure   TSynthPatch.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    var
      i  : Integer;
    begin
      try
        if aPrefix = ''
        then begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherInfo( Name, aCallback);
        end
        else begin
          for i := 0 to ModuleCount - 1
          do Module[ i].GatherInfo( aPrefix + '/' + Name, aCallback);
        end;
      except
        on E: Exception
        do KilledException( E);
      end;
    end;

    procedure   TSynthPatch.AcceptSamplePair1( aLeft, aRight: TSignal);
    begin
      FInputFifo1.Append( aLeft, aRight);
    end;

    procedure   TSynthPatch.AcceptSamplePair2( aLeft, aRight: TSignal);
    begin
      FInputFifo2.Append( aLeft, aRight);
    end;

    procedure   TSynthPatch.AcceptSamplePair3( aLeft, aRight: TSignal);
    begin
      FInputFifo3.Append( aLeft, aRight);
    end;

    procedure   TSynthPatch.AcceptSamplePair4( aLeft, aRight: TSignal);
    begin
      FInputFifo4.Append( aLeft, aRight);
    end;

    function    TSynthPatch.GetProfileHtml: TStringList;
    {$IFDEF PATCH_PROFILER}
    var
      i              : Integer;
      SumClocksFast  : Int64;
      SumClocksSlow  : Int64;
      SumSamplesFast : Int64;
      SumSamplesSlow : Int64;
      SumZips        : Int64;

      function Calc( Clocks, Samples: Int64): string; overload;
      begin
        if Samples = 0
        then Result := '---'
        else Result := Format( '%f', [ Clocks / Samples], AppLocale);
      end;

      function Calc( Clocks, Samples: TSignal): string; overload;
      begin
        if Samples = 0
        then Result := '---'
        else Result := Format( '%f', [ Clocks / Samples], AppLocale);
      end;

    {$ENDIF}
    begin
      Result := TStringList.Create;

      {$IFDEF PATCH_PROFILER}
      SumClocksFast  := 0;
      SumClocksSlow  := 0;
      SumSamplesFast := 0;
      SumSamplesSlow := 0;
      SumZips        := 0;

      Result.Add( '<!doctype html>');
      Result.Add( '<html>');
      Result.Add( '<head><style type="text/css">');
      Result.Add( 'table, td, th { border-style: solid; border-width: 1px; padding: 5px;}');
      Result.Add( '</style></head>');
      Result.Add( '<body><table><tbody>');
      Result.Add( '<tr><td>name</td><td>type</td><td>fast samples</td><td>slow samples</td><td>fast clocks</td><td>slow clocks</td><td>fast clocks / sample</td><td>slow clocks / sample</td></tr>');
      for i := 0 to ModuleCount - 1
      do begin
        Result.Add( FModules[ i].GetProfileInfoHtml);
        Inc( SumClocksFast , FModules[ i].TotalClocksFast);
        Inc( SumClocksSlow , FModules[ i].TotalClocksSlow);
        Inc( SumSamplesFast, FModules[ i].SamplesFast    );
        Inc( SumSamplesSlow, FModules[ i].SamplesSlow    );
        Inc( SumZips       , FModules[ i].TotalZips      );
      end;
      Result.Add( GetProfileInfoHtml);
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%s</td><td align="right">%s</td></tr>',
          [
            'total',
            'average module',
            SumSamplesFast,
            SumSamplesSlow,
            SumClocksFast,
            SumClocksSlow,
            Calc( SumClocksFast, SumSamplesFast),
            Calc( SumClocksSlow, SumSamplesSlow * Control_Decimation)
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%s</td><td></td><td></td></tr>',
          [
            'dezippers',
            '&nbsp;',
            SamplesFast,
            SamplesSlow,
            SumZips,
            Calc( SumZips, SamplesFast)
          ],
          AppLocale
        )
      );
      if ModuleCount > 0
      then begin
        Result.Add(
          Format(
            '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%s</td><td></td><td></td></tr>',
            [
              'dezippers / module',
              '&nbsp;',
              SamplesFast,
              SamplesSlow,
              Round( SumZips / ModuleCount),
              Calc( SumZips, SamplesFast * ModuleCount)
            ],
            AppLocale
          )
        );
      end;
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%s</td><td></td><td></td><td></td></tr>',
          [
            'wires',
            '&nbsp;',
            SamplesFast,
            FTotalWireClocks,
            Calc( FTotalWireClocks, SamplesFast)
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr> <td>%s</td> <td>%s</td> <td align="right">%d</td> <td align="right">%.0f</td> <td align="right">%s</td><td></td><td></td><td></td></tr>',
          [
            'wire',
            'average',
            SamplesFast,
            FTotalWireClocks / ConnectionCount,
            Calc( FTotalWireClocks / ConnectionCount, SamplesFast)
          ],
          AppLocale
        )
      );
      Result.Add( '<tr><td>name</td><td>type</td><td>fast samples</td><td>slow samples</td><td>fast clocks</td><td>slow clocks</td><td>fast clocks / sample</td><td>slow clocks / sample</td></tr>');
      Result.Add( '</tbody></table></body></html>');
      {$ENDIF}
    end;

    function    TSynthPatch.IsSamePatch( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then begin
        Result :=
          SameModules    ( aPatch) and
          SameConnections( aPatch)
      end;
    end;

    function    TSynthPatch.SameModules( const aPatch: TSynthPatch): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      if Assigned( aPatch)
      then begin
        if ModuleCount = aPatch.ModuleCount
        then begin
          Result := True;
          for i := 0 to ModuleCount - 1
          do begin
            if not Module[ i].IsSameModule( aPatch.Module[ i])
            then begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;

    function    TSynthPatch.SameConnections( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then Result := FConnections.IsSameConnections( aPatch.FConnections);
    end;

    procedure   TSynthPatch.AcceptMidi( const aMsg: TMidiMessage); // overide;
    var
      i : Integer;
    begin
      if FIsMidi
      then begin
        for i := 0 to ModuleCount - 1
        do begin
          if Module[ i].FIsMidi
          then Module[ i].AcceptMidi( aMsg);
        end;
      end;
    end;

    procedure   TSynthPatch.SendMidi( const aMsg: TMidiMessage); // override;
    begin
      if Assigned( FOnSendMidi)
      then FOnSendMidi( Self, aMsg);
    end;

    procedure   TSynthPatch.SampleRateChanged; // override;
    // Notification for sample rate change, the actual values are in global variables
    // System_Rate for the audio rate and Control_Rate for the control rate.
    // Modules caching sample rate dependent data should recalculate such data in an
    // overriden version of this procedure.
    var
      i : Integer;
    begin
      // Notify all sub-modules.
      for i := 0 to ModuleCount - 1
      do Module[ i].SampleRateChanged;
    end;

    procedure   TSynthPatch.ClearOverUnderFlows;
    begin
      if Assigned( FInputFifo1)
      then FInputFifo1.ClearOverUnderFlows;
      if Assigned( FInputFifo2)
      then FInputFifo1.ClearOverUnderFlows;
      if Assigned( FInputFifo3)
      then FInputFifo1.ClearOverUnderFlows;
      if Assigned( FInputFifo4)
      then FInputFifo1.ClearOverUnderFlows;
    end;





end.
